from datetime import datetime

from loguru import logger
from sqlalchemy import Column, DateTime, String, Unicode

from flexget import db_schema, plugin
from flexget.components.sites.utils import normalize_unicode
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event
from flexget.manager import Session
from flexget.plugin import PluginError
from flexget.utils.database import json_synonym
from flexget.utils.requests import TokenBucketLimiter
from flexget.utils.tools import parse_filesize

DETECT_2FA = "Authenticator Code", "TOTP code"
logger = logger.bind(name='gazelle')
Base = db_schema.versioned_base('gazelle_session', 0)


class GazelleSession(Base):
    __tablename__ = 'gazelle_session'

    username = Column(Unicode, primary_key=True)
    base_url = Column(String, primary_key=True)

    authkey = Column(String)
    passkey = Column(String)
    _cookies = Column('cookie', Unicode)
    cookies = json_synonym('_cookies')
    expires = Column(DateTime)


class InputGazelle:
    """A generic plugin that searches a Gazelle-based website

    Limited functionality but should work for almost all of them.
    """

    def __init__(self):
        """Set up a plugin that only has the ability to do about basic search"""
        self.base_url = None

        # Aliases for config -> api params
        # Extended in subclasses
        self.aliases = {"search": "searchstr"}

        # API parameters
        # None means a raw value entry (no validation other than schema)
        # A dict means an enum with a config -> api mapping
        # A list is an enum with no mapping
        # Extended in subclasses
        self.params = {"searchstr": None}

    @property
    def schema(self):
        """The schema of the plugin

        Subclasses should extend this to implement more params
        """
        schema = {
            'type': 'object',
            'properties': {
                'base_url': {'type': 'string'},
                'username': {'type': 'string'},
                'password': {'type': 'string'},
                'max_pages': {'type': 'integer'},
                'search': {'type': 'string'},
            },
            'required': ['username', 'password'],
            'additionalProperties': False,
        }
        # base_url is only required if the subclass doesn't set it
        if not self.base_url:
            schema['required'].append('base_url')
        return schema

    def _key(self, key):
        """Gets the API key name from the entered key"""
        if key in self.aliases:
            return self.aliases[key]
        return key

    def _opts(self, key):
        """Gets the options for the specified key"""
        return self.params[self._key(key)]

    def _getval(self, key, val):
        """Gets the value for the specified key based on a config option"""
        opts = self._opts(key)
        if isinstance(opts, dict):
            # Translate the input value to the API value
            # The str cast converts bools to 'True'/'False' for use as keys
            # This allows for options that have True/False/Other values
            return opts[str(val)]
        elif isinstance(val, list):
            # Fix yaml parser making a list out of a string
            return ",".join(val)
        return val

    def params_from_config(self, config):
        """Filter params and map config values -> api values"""
        ret = {}
        for k, v in config.items():
            key = self._key(k)
            if key in self.params:
                ret[key] = self._getval(k, v)
        return ret

    def setup(self, task, config):
        """Set up a session and log in"""
        self._session = task.requests
        base_url = config.get('base_url', "").rstrip("/")
        if base_url:
            if self.base_url and self.base_url != base_url:
                logger.warning(
                    'Using plugin designed for {} on {} - things may break',
                    self.base_url,
                    base_url,
                )
            self.base_url = base_url

        if not self.base_url:
            raise PluginError("No 'base_url' configured")

        # Any more than 5 pages is probably way too loose of a search
        self.max_pages = config.get('max_pages', 5)

        # The consistent request limiting rule seems to be:
        # "Refrain from making more than five (5) requests every ten (10) seconds"
        self._session.add_domain_limiter(TokenBucketLimiter(self.base_url, 2, '2 seconds'))

        self.username = config['username']
        self.password = config['password']

        # Login
        self.authenticate()

        # Logged in successfully, it's ok if nothing matches now
        task.no_entries_ok = True

    def resume_session(self):
        """Resume an existing session from the datebase

        Return True on successful recovery, False otherwise
        """
        logger.debug("Attempting to find an existing session in the DB")
        with Session() as session:
            db_session = (
                session.query(GazelleSession)
                .filter(
                    GazelleSession.base_url == self.base_url,
                    GazelleSession.username == self.username,
                )
                .one_or_none()
            )
            if db_session and db_session.expires and db_session.expires >= datetime.utcnow():
                # Found a valid session in the DB - use it
                self._session.cookies.update(db_session.cookies)
                self.authkey = db_session.authkey
                self.passkey = db_session.passkey
                return True
        return False

    def save_current_session(self):
        """Store the current session in the database so it can be resumed later"""
        logger.debug("Storing session info in the DB")
        with Session() as session:
            expires = None
            for c in self._session.cookies:
                if c.name == "session":
                    expires = datetime.utcfromtimestamp(c.expires)
            db_session = GazelleSession(
                username=self.username,
                base_url=self.base_url,
                cookies=dict(self._session.cookies),
                expires=expires,
                authkey=self.authkey,
                passkey=self.passkey,
            )
            session.merge(db_session)

    def authenticate(self, force=False):
        """Log in and store auth data from the server

        Adapted from https://github.com/isaaczafuta/whatapi
        """
        # clean slate before creating/restoring cookies
        self._session.cookies.clear()

        if not force and self.resume_session():
            logger.info('Logged into {} using cached session', self.base_url)
            return

        # Forcing a re-login or no session in DB - log in using provided creds
        url = "{}/login.php".format(self.base_url)
        data = {'username': self.username, 'password': self.password, 'keeplogged': 1}
        r = self._session.post(url, data=data, allow_redirects=False, raise_status=True)
        if not r.is_redirect or r.next.url != "{}/index.php".format(self.base_url):
            msg = "Failed to log into {}".format(self.base_url)
            for otp_text in DETECT_2FA:
                # TODO: Find a better signal that 2FA is enabled
                if otp_text in r.text:
                    msg += " - Accounts using 2FA are currently not supported"
                    break
            raise PluginError(msg)

        account_info = self.request(no_login=True, action='index')
        self.authkey = account_info['authkey']
        self.passkey = account_info['passkey']
        logger.info('Logged in to {}', self.base_url)

        # Store the session so we can resume it later
        self.save_current_session()

    def request(self, no_login=False, **params):
        """Make an AJAX request to the API

        If `no_login` is True, logging in will not be attempted if the request
        is redirected to the login page.

        Adapted from https://github.com/isaaczafuta/whatapi
        """
        if 'action' not in params:
            raise ValueError("An 'action' is required when making a request")

        ajaxpage = "{}/ajax.php".format(self.base_url)
        r = self._session.get(ajaxpage, params=params, allow_redirects=False, raise_status=True)
        if not no_login and r.is_redirect and r.next.url == "{}/login.php".format(self.base_url):
            logger.warning("Redirected to login page, reauthenticating and trying again")
            self.authenticate(force=True)
            return self.request(no_login=True, **params)

        if r.status_code != 200:
            raise PluginError("{} returned a non-200 status code".format(self.base_url))

        try:
            json_response = r.json()
            if json_response['status'] != "success":

                # Try to deal with errors returned by the API
                error = json_response.get('error', json_response.get('status'))
                if not error or error == "failure":
                    error = json_response.get('response', str(json_response))

                raise PluginError(
                    "{} gave a failure response of '{}'".format(self.base_url, error)
                )
            return json_response['response']
        except (ValueError, TypeError, KeyError):
            raise PluginError("{} returned an invalid response".format(self.base_url))

    def search_results(self, params):
        """Generator that yields search results"""
        page = 1
        pages = None
        while page <= self.max_pages:
            if pages and page >= pages:
                break

            logger.debug('Attempting to get page {} of search results', page)
            result = self.request(action='browse', page=page, **params)
            if not result['results']:
                break
            for x in result['results']:
                yield x

            pages = result.get('pages', pages)
            page += 1

        if page > self.max_pages:
            logger.warning('Stopped after {} pages (out of {} total pages)', self.max_pages, pages)

    def get_entries(self, search_results):
        """Generator that yields Entry objects from search results"""
        for result in search_results:
            # Get basic information on the release
            info = dict((k, result[k]) for k in ('groupId', 'groupName'))

            # Releases can have multiple download options
            for tor in result['torrents']:
                temp = info.copy()
                temp['torrentId'] = tor['torrentId']

                yield Entry(
                    title="{groupName} ({groupId} - {torrentId}).torrent".format(**temp),
                    url="{}/torrents.php?action=download&id={}&authkey={}&torrent_pass={}"
                    "".format(self.base_url, temp['torrentId'], self.authkey, self.passkey),
                    torrent_seeds=tor['seeders'],
                    torrent_leeches=tor['leechers'],
                    # Size is returned in bytes
                    content_size=parse_filesize(str(tor['size']) + "b"),
                )

    @plugin.internet(logger)
    def search(self, task, entry, config):
        """Search interface"""
        self.setup(task, config)

        entries = set()
        params = self.params_from_config(config)
        for search_string in entry.get('search_strings', [entry['title']]):
            query = normalize_unicode(search_string)
            params[self._key('search')] = query
            entries.update(self.get_entries(self.search_results(params)))
        return entries

    @plugin.internet(logger)
    def on_task_input(self, task, config):
        """Task input interface"""
        self.setup(task, config)

        params = self.params_from_config(config)
        return self.get_entries(self.search_results(params))


class InputGazelleMusic(InputGazelle):
    """A plugin that searches a Gazelle-based music website

    Based on https://github.com/WhatCD/Gazelle since it's the starting point of
    all Gazelle-based music sites.
    """

    def __init__(self):
        """Set up the majority of parameters that these sites support"""
        super().__init__()

        self.aliases.update(
            {
                "artist": "artistname",
                "album": "groupname",
                "leech_type": "freetorrent",
                "release_type": "releasetype",
                "tags": "taglist",
                "tag_type": "tags_type",
                "log": "haslog",
            }
        )

        self.params.update(
            {
                "taglist": None,
                "artistname": None,
                "groupname": None,
                "year": None,
                "tags_type": {"any": 0, "all": 1},
                "encoding": [
                    "192",
                    "APS (VBR)",
                    "V2 (VBR)",
                    "V1 (VBR)",
                    "256",
                    "APX (VBR)",
                    "V0 (VBR)",
                    "q8.x (VBR)",
                    "320",
                    "Lossless",
                    "24bit Lossless",
                    "Other",
                ],
                "format": ["MP3", "FLAC", "Ogg Vorbis", "AAC", "AC3", "DTS"],
                "media": ["CD", "DVD", "Vinyl", "Soundboard", "SACD", "DAT", "Cassette", "WEB"],
                "releasetype": {
                    "album": 1,
                    "soundtrack": 3,
                    "EP": 5,
                    "anthology": 6,
                    "compilation": 7,
                    "single": 9,
                    "live album": 11,
                    "remix": 13,
                    "bootleg": 14,
                    "interview": 15,
                    "mixtape": 16,
                    "unknown": 21,
                },
                "haslog": {"False": 0, "True": 1, "100%": 100, "<100%": -1},
                "freetorrent": {"freeleech": 1, "neutral": 2, "either": 3, "normal": 0},
                "hascue": {"False": 0, "True": 1},
                "scene": {"False": 0, "True": 1},
                "vanityhouse": {"False": 0, "True": 1},
            }
        )

    @property
    def schema(self):
        """The schema of the plugin

        Extends the super's schema
        """
        schema = super().schema
        schema['properties'].update(
            {
                'artist': {'type': 'string'},
                'album': {'type': 'string'},
                'year': {'type': ['string', 'integer']},
                'tags': one_or_more({'type': 'string'}),
                'tag_type': {'type': 'string', 'enum': list(self._opts('tag_type').keys())},
                'encoding': {'type': 'string', 'enum': self._opts('encoding')},
                'format': {'type': 'string', 'enum': self._opts('format')},
                'media': {'type': 'string', 'enum': self._opts('media')},
                'release_type': {
                    'type': 'string',
                    'enum': list(self._opts('release_type').keys()),
                },
                'log': {
                    'oneOf': [
                        {'type': 'string', 'enum': list(self._opts('log').keys())},
                        {'type': 'boolean'},
                    ]
                },
                'leech_type': {'type': 'string', 'enum': list(self._opts('leech_type').keys())},
                'hascue': {'type': 'boolean'},
                'scene': {'type': 'boolean'},
                'vanityhouse': {'type': 'boolean'},
            }
        )
        return schema

    def get_entries(self, search_results):
        """Generator that yields Entry objects from search results"""
        for result in search_results:
            # Get basic information on the release
            info = dict((k, result[k]) for k in ('artist', 'groupName', 'groupYear'))

            # Releases can have multiple download options
            for tor in result['torrents']:
                temp = info.copy()
                temp.update(
                    dict((k, tor[k]) for k in ('media', 'encoding', 'format', 'torrentId'))
                )

                yield Entry(
                    title="{artist} - {groupName} - {groupYear} "
                    "({media} - {format} - {encoding})-{torrentId}.torrent".format(**temp),
                    url="{}/torrents.php?action=download&id={}&authkey={}&torrent_pass={}"
                    "".format(self.base_url, temp['torrentId'], self.authkey, self.passkey),
                    torrent_seeds=tor['seeders'],
                    torrent_leeches=tor['leechers'],
                    # Size is returned in bytes
                    content_size=parse_filesize(str(tor['size']) + "b"),
                )


class InputRedacted(InputGazelleMusic):
    """A plugin that searches RED"""

    def __init__(self):
        """Set up custom base_url and parameters"""
        super().__init__()
        self.base_url = "https://redacted.ch"

        self.params['encoding'].remove("q8.x (VBR)")
        self.params['format'].remove("Ogg Vorbis")
        self.params['media'].append("Blu-ray")
        self.params['releasetype'].update({"demo": 17, "concert recording": 18, "dj mix": 19})


class InputNotWhat(InputGazelleMusic):
    """A plugin that searches NWCD"""

    def __init__(self):
        """Set up custom base_url and parameters"""
        super().__init__()
        self.base_url = "https://notwhat.cd"

        self.params['media'].extend(['Blu-ray', 'Unknown'])
        self.params['releasetype'].update({'demo': 22, 'dj mix': 23, 'concert recording': 24})
        self.params['haslog'].update(
            {
                "gold": 102,
                "silver": 101,
                "gold/silver": 100,
                "lineage": -5,
                "unscored": -1,
                "missing lineage": -6,
                "missing dr score": -7,
                "missing sample rate": -8,
                "missing description": -9,
            }
        )


@event('plugin.register')
def register_plugin():
    plugin.register(InputGazelle, 'gazelle', interfaces=['task', 'search'], api_ver=2)
    plugin.register(InputGazelleMusic, 'gazellemusic', interfaces=['task', 'search'], api_ver=2)
    plugin.register(InputRedacted, 'redacted', interfaces=['task', 'search'], api_ver=2)
    plugin.register(InputNotWhat, 'notwhatcd', interfaces=['task', 'search'], api_ver=2)



from loguru import logger
from requests.exceptions import RequestException

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.plugin import PluginWarning
from flexget.utils.requests import Session as RequestSession
from flexget.utils.requests import TimedLimiter

plugin_name = 'pushalot'
logger = logger.bind(name=plugin_name)

PUSHALOT_URL = 'https://pushalot.com/api/sendmessage'

requests = RequestSession(max_retries=3)
requests.add_domain_limiter(TimedLimiter('pushalot.com', '5 seconds'))


class PushalotNotifier:
    """
    Example::

      notify:
        entries:
          via:
            - pushalot:
                token: <string> Authorization token (can also be a list of tokens) - Required
                link: <string> (default: '{{imdb_url}}')
                linktitle: <string> (default: (none))
                important: <boolean> (default is False)
                silent: <boolean< (default is False)
                image: <string> (default: (none))
                source: <string> (default is 'FlexGet')
                timetolive: <integer>
    """

    schema = {
        'type': 'object',
        'properties': {
            'api_key': one_or_more({'type': 'string'}),
            'url': {'type': 'string'},
            'url_title': {'type': 'string'},
            'important': {'type': 'boolean', 'default': False},
            'silent': {'type': 'boolean', 'default': False},
            'image': {'type': 'string'},
            'source': {'type': 'string', 'default': 'FlexGet'},
            'timetolive': {'type': 'integer', 'maximum': 43200, 'minimum': 0},
        },
        'required': ['api_key'],
        'additionalProperties': False,
    }

    def notify(self, title, message, config):
        """
        Send a Pushalot notification
        """
        notification = {
            'Title': title,
            'Body': message,
            'LinkTitle': config.get('url_title'),
            'Link': config.get('url'),
            'IsImportant': config.get('important'),
            'IsSilent': config.get('silent'),
            'Image': config.get('image'),
            'Source': config.get('source'),
            'TimeToLive': config.get('timetolive'),
        }

        if not isinstance(config['api_key'], list):
            config['api_key'] = [config['api_key']]

        for key in config['api_key']:
            notification['AuthorizationToken'] = key
            try:
                requests.post(PUSHALOT_URL, json=notification)
            except RequestException as e:
                raise PluginWarning(repr(e))


@event('plugin.register')
def register_plugin():
    plugin.register(PushalotNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



import os
import random
from collections import namedtuple

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more, parse_percent, parse_size
from flexget.event import event

logger = logger.bind(name='path_by_space')

disk_stats_tuple = namedtuple(
    'disk_stats',
    ['path', 'free_bytes', 'used_bytes', 'total_bytes', 'free_percent', 'used_percent'],
)


def os_disk_stats(folder):
    """Return drive free, used and total bytes"""
    if os.name == 'nt':
        import ctypes

        free_bytes = ctypes.c_ulonglong(0)
        total_bytes = ctypes.c_ulonglong(0)
        ctypes.windll.kernel32.GetDiskFreeSpaceExW(
            ctypes.c_wchar_p(folder), None, ctypes.pointer(total_bytes), ctypes.pointer(free_bytes)
        )

        return free_bytes.value, total_bytes.value
    else:
        stats = os.statvfs(folder)
        return stats.f_bavail * stats.f_frsize, stats.f_blocks * stats.f_frsize


def disk_stats(folder):
    free_bytes, total_bytes = os_disk_stats(folder)
    used_bytes = total_bytes - free_bytes
    free_percent = 0.0 if total_bytes == 0 else 100 * free_bytes / total_bytes
    used_percent = 0.0 if total_bytes == 0 else 100 * used_bytes / total_bytes

    return disk_stats_tuple(
        folder, free_bytes, used_bytes, total_bytes, free_percent, used_percent
    )


def _path_selector(paths, within, stat_attr):
    paths_stats = [disk_stats(path) for path in paths]

    # Sort paths by key
    paths_stats.sort(key=lambda p: getattr(p, stat_attr), reverse=True)

    valid_paths = [paths_stats[0].path]

    if within > 0:
        valid_paths.extend(
            [
                path_stat.path
                for path_stat in paths_stats[1:]
                if abs(getattr(path_stat, stat_attr) - getattr(paths_stats[0], stat_attr))
                <= within
            ]
        )

    return random.choice(valid_paths)


def select_most_free(paths, within):
    return _path_selector(paths, within, 'free_bytes')


def select_most_used(paths, within):
    return _path_selector(paths, within, 'used_bytes')


def select_most_free_percent(paths, within):
    return _path_selector(paths, within, 'free_percent')


def select_most_used_percent(paths, within):
    return _path_selector(paths, within, 'used_percent')


selector_map = {
    'most_free': select_most_free,
    'most_used': select_most_used,
    'most_free_percent': select_most_free_percent,
    'most_used_percent': select_most_used_percent,
}


class PluginPathBySpace:
    """Allows setting a field to a folder based on it's space

    Path will be selected at random if multiple paths match the within

    Example:

    path_by_space:
      select: most_free_percent # or most_free, most_used, most_used_percent, has_free
      within: 9000 # within in MB or percent.
      paths:
        - /drive1/
        - /drive2/
        - /drive3/
    """

    schema = {
        'type': 'object',
        'properties': {
            'select': {'type': 'string', 'enum': list(selector_map.keys())},
            'to_field': {'type': 'string', 'default': 'path'},
            'paths': one_or_more({'type': 'string', 'format': 'path'}),
            'within': {
                'oneOf': [
                    {'type': 'string', 'format': 'size'},
                    {'type': 'string', 'format': 'percent'},
                ]
            },
        },
        'required': ['paths', 'select'],
        'additionalProperties': False,
    }

    @plugin.priority(250)  # run before other plugins
    def on_task_metainfo(self, task, config):
        selector = selector_map[config['select']]

        # Convert within to bytes (int) or percent (float)
        within = config.get('within')
        if isinstance(within, str) and '%' in within:
            within = parse_percent(within)
        else:
            within = parse_size(within)

        path = selector(config['paths'], within=within)

        if path:
            logger.debug('Path {} selected due to ({})', path, config['select'])

            for entry in task.all_entries:
                entry[config['to_field']] = path
        else:
            task.abort('Unable to select a path based on %s' % config['select'])
            return


@event('plugin.register')
def register_plugin():
    plugin.register(PluginPathBySpace, 'path_by_space', api_ver=2)



import re
import time
from datetime import datetime
from urllib.parse import urlparse

from loguru import logger

from flexget import plugin
from flexget.components.sites.urlrewriting import UrlRewritingError
from flexget.components.sites.utils import normalize_unicode, torrent_availability
from flexget.entry import Entry
from flexget.event import event

logger = logger.bind(name='solidtorrents')

URL = 'https://solidtorrents.net/'

CATEGORIES = (
    "all",
    "Audio",
    "Video",
    "Image",
    "Document",
    "eBook",
    "Program",
    "Android",
    "Archive",
    "Diskimage",
    "Sourcecode",
    "Database",
)


SORT = ("seeders", "leechers", "downloads", "date", "size")


class UrlRewriteSolidTorrents:
    """SolidTorrents urlrewriter."""

    schema = {
        'oneOf': [
            {'type': 'boolean'},
            {
                'type': 'object',
                'properties': {
                    'url': {'type': 'string', 'default': URL, 'format': 'url'},
                    'category': {'type': 'string', 'enum': list(CATEGORIES)},
                    'sort_by': {'type': 'string', 'enum': list(SORT)},
                    'reverse': {'type': 'boolean'},
                    'remove_potentially_unsafe': {'type': 'boolean'},
                },
                'additionalProperties': False,
            },
        ]
    }

    url = None

    def __init__(self):
        self.set_urls(URL)

    def on_task_start(self, task, config=None):
        if not isinstance(config, dict):
            config = {}
        self.set_urls(config.get('url', URL))

    def set_urls(self, url):
        url = url.rstrip('/')
        if self.url != url:
            self.url = url
            parsed_url = urlparse(url)
            escaped_url_scheme = re.escape(parsed_url.scheme)
            escaped_url_netloc = re.escape(parsed_url.netloc)
            self.url_match = re.compile(
                fr'^{escaped_url_scheme}://{escaped_url_netloc}/view/[^/]+/([a-zA-Z0-9]{24})$'
            )
            self.url_search = re.compile(
                fr'^{escaped_url_scheme}://{escaped_url_netloc}/search\?q=.*$'
            )

    # urlrewriter API
    def url_rewritable(self, task, entry):
        url = entry['url']
        if url.endswith('.torrent'):
            return False
        return bool(self.url_match.match(url))

    # urlrewriter API
    def url_rewrite(self, task, entry):
        if 'url' not in entry:
            logger.error("Didn't actually get a URL...")
        else:
            logger.debug('Got the URL: {}', entry['url'])
        if self.url_search.match(entry['url']):
            # use search
            results = self.search(task, entry)
            if not results:
                raise UrlRewritingError("No search results found")
            # TODO: Close matching was taken out of search methods, this may need to be fixed to be more picky
            entry['url'] = results[0]['url']
        else:
            torrent_id = self.url_match(entry['url']).group(1)
            url = f"{self.url}/api/v1/torrents/{torrent_id}"
            logger.debug('Getting info for torrent ID {}', torrent_id)
            json_result = task.requests.get(url).json()
            # if json_result['error'] == '404':
            if not 'result' in json_result:
                raise UrlRewritingError("Torrent with ID %s does not exist." % torrent_id)
            entry['url'] = json_result['result']['rating']['magnet']

    @plugin.internet(logger)
    def search(self, task, entry, config=None):
        """
        Search for name from solidtorrents.
        """
        if not isinstance(config, dict):
            config = {}
        self.set_urls(config.get('url', URL))
        sort = config.get('sort_by', 'seeders')
        reverse = bool(config.get('reverse', 'False'))
        remove_potentially_unsafe = bool(config.get('remove_potentially_unsafe', 'True'))
        category = config.get('category', 'all')

        entries = list()
        for search_string in entry.get('search_strings', [entry['title']]):
            # query = normalize_unicode(search_string)
            params = {
                'q': search_string,
                'cat': category,
                'sort': sort,
                'fuv': remove_potentially_unsafe,
            }
            url = f"{self.url}/api/v1/search"
            json_results = task.requests.get(url, params=params).json()
            if not json_results:
                raise plugin.PluginError(
                    "Error while searching solidtorrents for %s.", search_string
                )
            if not 'results' in json_results:
                logger.info(
                    "No result founds while searching solidtorrents for %s.", search_string
                )
                continue
            for result in json_results['results']:
                entry = self.json_to_entry(result)
                entries.append(entry)

        return reversed(entries) if reverse else entries

    # convert a single json result to an Entry
    def json_to_entry(self, json_result: dict) -> Entry:
        entry = Entry()
        entry['title'] = json_result['title']
        entry['torrent_seeds'] = int(json_result['swarm']['seeders'])
        entry['torrent_leeches'] = int(json_result['swarm']['leechers'])
        entry['torrent_timestamp'] = int(
            time.mktime(
                datetime.strptime(json_result['imported'], '%Y-%m-%dT%H:%M:%S.%fZ').timetuple()
            )
        )
        entry['torrent_availability'] = torrent_availability(
            entry['torrent_seeds'], entry['torrent_leeches']
        )
        entry['content_size'] = int(
            round(int(json_result['size']) / (1024 * 1024))
        )  # content_size is in MiB
        entry['torrent_info_hash'] = json_result['infohash']
        entry['url'] = json_result['magnet']
        return entry


@event('plugin.register')
def register_plugin():
    plugin.register(
        UrlRewriteSolidTorrents,
        'solidtorrents',
        interfaces=['urlrewriter', 'search', 'task'],
        api_ver=2,
    )



from datetime import datetime, timedelta

from loguru import logger
from sqlalchemy import Boolean, Column, DateTime, Float, Integer, Table, Text, Unicode
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relation
from sqlalchemy.schema import ForeignKey

from flexget import db_schema
from flexget.utils import requests
from flexget.utils.database import Session, json_synonym, text_date_synonym, with_session
from flexget.utils.simple_persistence import SimplePersistence
from flexget.utils.tools import chunked, split_title_year

logger = logger.bind(name='api_tvdb')
Base = db_schema.versioned_base('api_tvdb', 7)

persist = SimplePersistence('api_tvdb')

SEARCH_RESULT_EXPIRATION_DAYS = 3


class TVDBRequest:
    # This is a FlexGet API key
    API_KEY = '4D297D8CFDE0E105'
    BASE_URL = 'https://api.thetvdb.com/'
    BANNER_URL = 'http://thetvdb.com/banners/'

    def __init__(self, username=None, account_id=None, api_key=None):
        self.username = username
        self.account_id = account_id
        self.api_key = api_key
        self.auth_key = self.username.lower() if self.username else 'default'

    def get_auth_token(self, refresh=False):
        with Session() as session:
            auth_token = session.query(TVDBTokens).filter(TVDBTokens.name == self.auth_key).first()
            if not auth_token:
                auth_token = TVDBTokens()
                auth_token.name = self.auth_key

            if refresh or auth_token.has_expired():
                data = {'apikey': TVDBRequest.API_KEY}
                if self.username:
                    data['username'] = self.username
                if self.account_id:
                    data['userkey'] = self.account_id
                if self.api_key:
                    data['apikey'] = self.api_key

                logger.debug(
                    'Authenticating to TheTVDB with {}',
                    self.username if self.username else 'api_key',
                )

                auth_token.token = (
                    requests.post(TVDBRequest.BASE_URL + 'login', json=data).json().get('token')
                )
                auth_token.refreshed = datetime.now()
                auth_token = session.merge(auth_token)

            return auth_token.token

    def _request(self, method, endpoint, **params):
        url = TVDBRequest.BASE_URL + endpoint
        language = params.pop('language', 'en')
        headers = {'Authorization': f'Bearer {self.get_auth_token()}', 'Accept-Language': language}
        data = params.pop('data', None)

        result = requests.request(
            method, url, params=params, headers=headers, raise_status=False, json=data
        )
        if result.status_code == 401:
            logger.debug('Auth token expired, refreshing')
            headers['Authorization'] = f'Bearer {self.get_auth_token(refresh=True)}'
            result = requests.request(
                method, url, params=params, headers=headers, raise_status=False, json=data
            )
        result.raise_for_status()
        result = result.json()

        if result.get('errors'):
            logger.debug('Result contains errors: {}', result['errors'])
            # a hack to make sure it doesn't raise exception on a simple invalidLanguage. This is because tvdb
            # has a tendency to contain bad data and randomly return this error for no reason
            if len(result['errors']) > 1 or 'invalidLanguage' not in result['errors']:
                raise LookupError(f'Error processing request on tvdb: {result["errors"]}')

        return result

    def get(self, endpoint, **params):
        result = self._request('get', endpoint, **params)
        return result.get('data')

    def post(self, endpoint, **params):
        result = self._request('post', endpoint, **params)
        return result.get('data')

    def put(self, endpoint, **params):
        result = self._request('put', endpoint, **params)
        return result.get('data')

    def delete(self, endpoint, **params):
        return self._request('delete', endpoint, **params)


@db_schema.upgrade('api_tvdb')
def upgrade(ver, session):
    if ver is None or ver <= 6:
        raise db_schema.UpgradeImpossible
    return ver


# association tables
genres_table = Table(
    'tvdb_series_genres',
    Base.metadata,
    Column('series_id', Integer, ForeignKey('tvdb_series.id')),
    Column('genre_id', Integer, ForeignKey('tvdb_genres.id')),
)
Base.register_table(genres_table)


class TVDBTokens(Base):
    __tablename__ = 'tvdb_tokens'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(Unicode)
    token = Column(Text)
    refreshed = Column(DateTime)

    def has_expired(self):
        if not self.token or not self.refreshed:
            return True

        seconds = (datetime.now() - self.refreshed).total_seconds()
        if seconds >= 86400:
            return True

        return False


class TVDBSeries(Base):
    __tablename__ = "tvdb_series"

    id = Column(Integer, primary_key=True, autoincrement=False)
    last_updated = Column(Integer)
    expired = Column(Boolean)
    name = Column(Unicode)
    language = Column(Unicode)
    rating = Column(Float)
    status = Column(Unicode)
    runtime = Column(Integer)
    airs_time = Column(Unicode)
    airs_dayofweek = Column(Unicode)
    content_rating = Column(Unicode)
    network = Column(Unicode)
    overview = Column(Unicode)
    imdb_id = Column(Unicode)
    zap2it_id = Column(Unicode)
    _banner = Column('banner', Unicode)

    _first_aired = Column('first_aired', DateTime)
    first_aired = text_date_synonym('_first_aired')
    _aliases = Column('aliases', Unicode)
    aliases = json_synonym('_aliases')
    _actors = Column('actors', Unicode)
    actors_list = json_synonym('_actors')
    _posters = Column('posters', Unicode)
    posters_list = json_synonym('_posters')

    _genres = relation('TVDBGenre', secondary=genres_table)
    genres = association_proxy('_genres', 'name')

    episodes = relation('TVDBEpisode', backref='series', cascade='all, delete, delete-orphan')

    def __init__(self, tvdb_id, language):
        """
        Looks up movie on tvdb and creates a new database model for it.
        These instances should only be added to a session via `session.merge`.
        """
        self.id = tvdb_id

        try:
            series = TVDBRequest().get(f'series/{self.id}', language=language)
        except requests.RequestException as e:
            raise LookupError(f'Error updating data from tvdb: {e}')

        self.language = language or 'en'
        self.last_updated = series['lastUpdated']
        self.name = series['seriesName']
        self.rating = float(series['siteRating']) if series['siteRating'] else 0.0
        self.status = series['status']
        self.runtime = int(series['runtime']) if series['runtime'] else 0
        self.airs_time = series['airsTime']
        self.airs_dayofweek = series['airsDayOfWeek']
        self.content_rating = series['rating']
        self.network = series['network']
        self.overview = series['overview']
        self.imdb_id = series['imdbId']
        self.zap2it_id = series['zap2itId']
        self.first_aired = series['firstAired']
        self.expired = False
        self.aliases = series['aliases']
        self._banner = series['banner']
        self._genres = [TVDBGenre(id=name) for name in series['genre']] if series['genre'] else []

        if not self.name:
            raise LookupError(
                f'Not possible to get name to series with id {self.id} in language \'{self.language}\''
            )

        if self.first_aired is None:
            logger.debug(
                'Falling back to getting first episode aired date for series {}', self.name
            )
            try:
                episode = TVDBRequest().get(
                    f'series/{self.id}/episodes/query?airedSeason=1&airedEpisode=1',
                    language=language,
                )
                self.first_aired = episode[0]['firstAired']
            except requests.RequestException:
                logger.error('Failed to get first episode for series {}', self.name)

        # Actors and Posters are lazy populated
        self._actors = None
        self._posters = None

    def __repr__(self):
        return f'<TVDBSeries name={self.name},tvdb_id={self.id}>'

    @property
    def banner(self):
        if self._banner:
            return TVDBRequest.BANNER_URL + self._banner

    @property
    def actors(self):
        return self.get_actors()

    @property
    def posters(self):
        return self.get_posters()

    def get_actors(self):
        if not self._actors:
            logger.debug('Looking up actors for series {}', self.name)
            try:
                actors_query = TVDBRequest().get(f'series/{self.id}/actors')
                self.actors_list = [a['name'] for a in actors_query] if actors_query else []
            except requests.RequestException as e:
                if None is not e.response and e.response.status_code == 404:
                    self.actors_list = []
                else:
                    raise LookupError(f'Error updating actors from tvdb: {e}')

        return self.actors_list

    def get_posters(self):
        if not self._posters:
            logger.debug('Getting top 5 posters for series {}', self.name)
            try:
                poster_main = TVDBRequest().get(f'series/{self.id}').get('poster')
                poster_query = TVDBRequest().get(
                    f'series/{self.id}/images/query', keyType='poster'
                )
                self.posters_list = [poster_main] if poster_main else []
                self.posters_list += (
                    [p['fileName'] for p in poster_query[:5]] if poster_query else []
                )
            except requests.RequestException as e:
                if None is not e.response and e.response.status_code == 404:
                    self.posters_list = []
                else:
                    raise LookupError(f'Error updating posters from tvdb: {e}')

        return [TVDBRequest.BANNER_URL + p for p in self.posters_list[:5]]

    def to_dict(self):
        return {
            'aliases': [a for a in self.aliases],
            'tvdb_id': self.id,
            'last_updated': datetime.fromtimestamp(self.last_updated).strftime(
                '%Y-%m-%d %H:%M:%S'
            ),
            'expired': self.expired,
            'series_name': self.name,
            'language': self.language,
            'rating': self.rating,
            'status': self.status,
            'runtime': self.runtime,
            'airs_time': self.airs_time,
            'airs_dayofweek': self.airs_dayofweek,
            'content_rating': self.content_rating,
            'network': self.network,
            'overview': self.overview,
            'imdb_id': self.imdb_id,
            'zap2it_id': self.zap2it_id,
            'banner': self.banner,
            'posters': self.posters,
            'genres': [g for g in self.genres],
            'first_aired': self.first_aired,
        }


class TVDBGenre(Base):
    __tablename__ = 'tvdb_genres'
    id = Column(Unicode, primary_key=True)

    @property
    def name(self):
        return self.id


class TVDBEpisode(Base):
    __tablename__ = 'tvdb_episodes'

    id = Column(Integer, primary_key=True, autoincrement=False)
    expired = Column(Boolean)
    last_updated = Column(Integer)
    season_number = Column(Integer)
    episode_number = Column(Integer)
    absolute_number = Column(Integer)
    name = Column(Unicode)
    overview = Column(Unicode)
    rating = Column(Float)
    director = Column(Unicode)
    _image = Column(Unicode)
    _first_aired = Column('firstaired', DateTime)
    first_aired = text_date_synonym('_first_aired')

    series_id = Column(Integer, ForeignKey('tvdb_series.id'), nullable=False)

    def __init__(self, series_id, ep_id, language=None):
        """
        Looks up movie on tvdb and creates a new database model for it.
        These instances should only be added to a session via `session.merge`.
        """
        self.series_id = series_id
        self.id = ep_id
        self.expired = False
        try:
            episode = TVDBRequest().get(f'episodes/{self.id}', language=language)
        except requests.RequestException as e:
            raise LookupError(f'Error updating data from tvdb: {e}')

        self.id = episode['id']
        self.last_updated = episode['lastUpdated']
        self.season_number = episode['airedSeason']
        self.episode_number = episode['airedEpisodeNumber']
        self.absolute_number = episode['absoluteNumber']
        self.name = episode['episodeName']
        self.overview = episode['overview']
        self.director = ', '.join(episode['directors'])
        self._image = episode['filename']
        self.rating = episode['siteRating']
        self.first_aired = episode['firstAired']

    def __repr__(self):
        return (
            f'<TVDBEpisode series={self.series.name},season={self.season_number},'
            f'episode={self.episode_number}>'
        )

    def to_dict(self):
        return {
            'id': self.id,
            'expired': self.expired,
            'last_update': self.last_updated,
            'season_number': self.season_number,
            'episode_number': self.episode_number,
            'absolute_number': self.absolute_number,
            'episode_name': self.name,
            'overview': self.overview,
            'director': self.director,
            'rating': self.rating,
            'image': self.image,
            'first_aired': self.first_aired,
            'series_id': self.series_id,
        }

    @property
    def image(self):
        if self._image:
            return TVDBRequest.BANNER_URL + self._image


class TVDBSearchResult(Base):
    __tablename__ = 'tvdb_search_results'

    id = Column(Integer, primary_key=True)
    search = Column(Unicode, nullable=False, unique=True)
    series_id = Column(Integer, ForeignKey('tvdb_series.id'), nullable=True)
    series = relation(TVDBSeries, backref='search_strings')

    def __init__(self, search, series_id=None, series=None):
        self.search = search.lower()
        if series_id:
            self.series_id = series_id
        if series:
            self.series = series


class TVDBSeriesSearchResult(Base):
    """
    This table will hold a single result that results from the /search/series endpoint,
    which return a series with a minimal set of parameters.
    """

    __tablename__ = 'tvdb_series_search_results'

    id = Column(Integer, primary_key=True, autoincrement=False)
    lookup_term = Column(Unicode)

    name = Column(Unicode)
    status = Column(Unicode)
    network = Column(Unicode)
    overview = Column(Unicode)

    _banner = Column('banner', Unicode)

    _first_aired = Column('first_aired', DateTime)
    first_aired = text_date_synonym('_first_aired')

    _aliases = Column('aliases', Unicode)
    aliases = json_synonym('_aliases')

    created_at = Column(DateTime)
    search_name = Column(Unicode)

    def __init__(self, series, lookup_term=None):
        self.lookup_term = lookup_term
        self.id = series['id']
        self.name = series['seriesName']
        self.first_aired = series['firstAired']
        self.network = series['network']
        self.overview = series['overview']
        self.status = series['status']
        self._banner = series['banner']
        self.aliases = series['aliases']
        self.created_at = datetime.now()

    @property
    def banner(self):
        if self._banner:
            return TVDBRequest.BANNER_URL + self._banner

    def to_dict(self):
        return {
            'aliases': [a for a in self.aliases],
            'banner': self.banner,
            'first_aired': self.first_aired,
            'tvdb_id': self.id,
            'network': self.network,
            'overview': self.overview,
            'series_name': self.name,
            'status': self.status,
        }

    @property
    def expired(self):
        logger.debug('checking series {} for expiration', self.name)
        if datetime.now() - self.created_at >= timedelta(days=SEARCH_RESULT_EXPIRATION_DAYS):
            logger.debug('series {} is expires, should re-fetch', self.name)
            return True
        logger.debug('series {} is not expired', self.name)
        return False


def find_series_id(name, language=None):
    """Looks up the tvdb id for a series"""
    try:
        series = TVDBRequest().get('search/series', name=name, language=language)
    except requests.RequestException as e:
        raise LookupError(f'Unable to get search results for {name}: {e}')

    name = name.lower()

    if not series:
        raise LookupError(f'No results found for {name}')

    # Cleanup results for sorting
    for s in series:
        if s['firstAired']:
            try:
                s['firstAired'] = datetime.strptime(s['firstAired'], "%Y-%m-%d")
            except ValueError:
                logger.warning(
                    'Invalid firstAired date "{}" when parsing series {} ',
                    s['firstAired'],
                    s['seriesName'],
                )
                s['firstAired'] = datetime(1970, 1, 1)
        else:
            s['firstAired'] = datetime(1970, 1, 1)

        s['names'] = [a.lower() for a in s.get('aliases')] if s.get('aliases') else []
        if s.get('seriesName'):
            s['names'].append(s.get('seriesName').lower())
        s['running'] = True if s['status'] == 'Continuing' else False

        for n in s['names']:
            # Exact matching by stripping our the year
            title = split_title_year(n)[0]
            if title not in s['names']:
                s['names'].append(title)

    # Sort by status, aired_date
    series = sorted(series, key=lambda x: (x['running'], x['firstAired']), reverse=True)

    for s in series:
        # Exact match
        if name in s['names']:
            return s['id']

    # If there is no exact match, pick the first result
    return series[0]['id']


def _update_search_strings(series, session, search=None):
    search_strings = series.search_strings
    aliases = [a.lower() for a in series.aliases] if series.aliases else []
    searches = [search.lower()] if search else []
    add = [series.name.lower()] + aliases + searches
    for name in set(add):
        if name not in search_strings:
            search_result = (
                session.query(TVDBSearchResult).filter(TVDBSearchResult.search == name).first()
            )
            if not search_result:
                search_result = TVDBSearchResult(search=name)
            search_result.series_id = series.id
            session.add(search_result)


@with_session
def lookup_series(name=None, tvdb_id=None, only_cached=False, session=None, language=None):
    """
    Look up information on a series. Will be returned from cache if available, and looked up online and cached if not.

    Either `name` or `tvdb_id` parameter are needed to specify the series.
    :param unicode name: Name of series.
    :param int tvdb_id: TVDb ID of series.
    :param bool only_cached: If True, will not cause an online lookup. LookupError will be raised if not available
        in the cache.
    :param session: An sqlalchemy session to be used to lookup and store to cache. Commit(s) may occur when passing in
        a session. If one is not supplied it will be created.
    :param language: Language abbreviation string to be sent to API

    :return: Instance of :class:`TVDBSeries` populated with series information. If session was not supplied, this will
        be a detached from the database, so relationships cannot be loaded.
    :raises: :class:`LookupError` if series cannot be looked up.
    """
    if not (name or tvdb_id):
        raise LookupError('No criteria specified for tvdb lookup')

    logger.debug("Looking up tvdb information for '{}'. TVDB ID: {}", name, tvdb_id)

    series = None

    def id_str():
        return f'<name={name},tvdb_id={tvdb_id}>'

    if tvdb_id:
        series = session.query(TVDBSeries).filter(TVDBSeries.id == tvdb_id).first()
    if not series and name:
        found = (
            session.query(TVDBSearchResult).filter(TVDBSearchResult.search == name.lower()).first()
        )
        if found and found.series:
            series = found.series
    if series:
        # Series found in cache, update if cache has expired.
        if not only_cached:
            mark_expired(session)
        if not only_cached and series.expired:
            logger.verbose('Data for {} has expired, refreshing from tvdb', series.name)
            try:
                updated_series = TVDBSeries(series.id, language)
                series = session.merge(updated_series)
                if series and series.name:
                    _update_search_strings(series, session, search=name)
            except LookupError as e:
                logger.warning(
                    'Error while updating from tvdb ({}), using cached data.', e.args[0]
                )
        else:
            logger.debug('Series {} information restored from cache.', id_str())
    else:
        if only_cached:
            raise LookupError(f'Series {id_str()} not found from cache')
        # There was no series found in the cache, do a lookup from tvdb
        logger.debug('Series {} not found in cache, looking up from tvdb.', id_str())
        if tvdb_id:
            series = session.merge(TVDBSeries(tvdb_id, language))
        elif name:
            tvdb_id = find_series_id(name, language=language)
            if tvdb_id:
                series = session.query(TVDBSeries).filter(TVDBSeries.id == tvdb_id).first()
                if not series:
                    series = session.merge(TVDBSeries(tvdb_id, language))
        if series and series.name:
            _update_search_strings(series, session, search=name)

    if not series:
        raise LookupError(f'No results found from tvdb for {id_str()}')
    if not series.name:
        raise LookupError('Tvdb result for series does not have a title.')

    return series


@with_session
def lookup_episode(
    name=None,
    season_number=None,
    episode_number=None,
    absolute_number=None,
    tvdb_id=None,
    first_aired=None,
    only_cached=False,
    session=None,
    language=None,
):
    """
    Look up information on an episode. Will be returned from cache if available, and looked up online and cached if not.

    Either `name` or `tvdb_id` parameter are needed to specify the series.
    Either `season_number` and `episode_number`, `absolute_number`, or `first_aired` are required to specify episode
     number.
    :param unicode name: Name of series episode belongs to.
    :param int tvdb_id: TVDb ID of series episode belongs to.
    :param int season_number: Season number of episode.
    :param int episode_number: Episode number of episode.
    :param int absolute_number: Absolute number of episode.
    :param date first_aired: Air date of episode. DateTime object.
    :param bool only_cached: If True, will not cause an online lookup. LookupError will be raised if not available
        in the cache.
    :param session: An sqlalchemy session to be used to lookup and store to cache. Commit(s) may occur when passing in
        a session. If one is not supplied it will be created, however if you need to access relationships you should
        pass one in.
    :param language: Language abbreviation string to be sent to API

    :return: Instance of :class:`TVDBEpisode` populated with series information.
    :raises: :class:`LookupError` if episode cannot be looked up.
    """
    # First make sure we have the series data
    series = lookup_series(name=name, tvdb_id=tvdb_id, only_cached=only_cached, session=session)

    if not series:
        raise LookupError(f'Series {name} ({tvdb_id}) not found from')

    ep_description = series.name
    query_params = {}
    episode = session.query(TVDBEpisode).filter(TVDBEpisode.series_id == series.id)

    if absolute_number is not None:
        episode = episode.filter(TVDBEpisode.absolute_number == absolute_number)
        query_params['absoluteNumber'] = absolute_number
        ep_description = f'{ep_description} absNo: {absolute_number}'

    if season_number is not None:
        episode = episode.filter(TVDBEpisode.season_number == season_number)
        query_params['airedSeason'] = season_number
        ep_description = f'{ep_description} s{season_number}'

    if episode_number is not None:
        episode = episode.filter(TVDBEpisode.episode_number == episode_number)
        query_params['airedEpisode'] = episode_number
        ep_description = f'{ep_description} e{episode_number}'

    if first_aired is not None:
        episode = episode.filter(TVDBEpisode.first_aired == first_aired)
        query_params['firstAired'] = datetime.strftime(first_aired, '%Y-%m-%d')
        ep_description = f'{ep_description} e{first_aired}'

    episode = episode.first()

    if episode:
        if episode.expired and not only_cached:
            logger.info('Data for {!r} has expired, refreshing from tvdb', episode)
            try:
                updated_episode = TVDBEpisode(series.id, episode.id)
                episode = session.merge(updated_episode)
            except LookupError as e:
                logger.warning('Error while updating from tvdb ({}), using cached data.', str(e))
        else:
            logger.debug('Using episode info for {} from cache.', ep_description)
    else:
        if only_cached:
            raise LookupError(f'Episode {ep_description} not found from cache')
        # There was no episode found in the cache, do a lookup from tvdb
        logger.debug('Episode {} not found in cache, looking up from tvdb.', ep_description)
        try:
            results = TVDBRequest().get(
                f'series/{series.id}/episodes/query', language=language, **query_params
            )
            if results:
                # Check if this episode id is already in our db
                episode = (
                    session.query(TVDBEpisode).filter(TVDBEpisode.id == results[0]['id']).first()
                )
                if not episode or (episode and episode.expired is not False):
                    updated_episode = TVDBEpisode(series.id, results[0]['id'], language=language)
                    episode = session.merge(updated_episode)

        except requests.RequestException as e:
            raise LookupError(f'Error looking up episode from TVDb ({e})')
    if episode:
        return episode
    else:
        raise LookupError(f'No results found for {ep_description}')


@with_session
def search_for_series(
    search_name=None, imdb_id=None, zap2it_id=None, force_search=None, session=None, language=None
):
    """
    Search IMDB using a an identifier, return a list of cached search results.
    One of `search_name`, `imdb_id` or `zap2it_id` is required.
    :param search_name: Name of search to use
    :param imdb_id: Search via IMDB ID
    :param zap2it_id: Search via zap2it ID
    :param force_search: Should search use cache
    :param session: Current DB session
    :return: A List of TVDBSeriesSearchResult objects
    """
    if imdb_id:
        lookup_term = imdb_id
        lookup_url = f'search/series?imdbId={imdb_id}'
    elif zap2it_id:
        lookup_term = zap2it_id
        lookup_url = f'search/series?zap2itId={zap2it_id}'
    elif search_name:
        lookup_term = search_name
        lookup_url = f'search/series?name={search_name}'
    else:
        raise LookupError('not enough parameters for lookup')
    series_search_results = []
    if not force_search:
        logger.debug('trying to fetch TVDB search results from DB')
        series_search_results = (
            session.query(TVDBSeriesSearchResult)
            .filter(TVDBSeriesSearchResult.lookup_term == lookup_term)
            .all()
        )

    if not series_search_results or any(series.expired for series in series_search_results):
        try:
            logger.debug('trying to fetch TVDB search results from TVDB')
            fetched_results = TVDBRequest().get(lookup_url, language=language)
        except requests.RequestException as e:
            raise LookupError(f'Error searching series from TVDb ({e})')
        series_search_results = [
            session.merge(TVDBSeriesSearchResult(series, lookup_term))
            for series in fetched_results
        ]
    if series_search_results:
        return series_search_results
    raise LookupError('No results found for series lookup')


def mark_expired(session):
    """Marks series and episodes that have expired since we cached them"""
    # Only get the expired list every hour
    last_check = persist.get('last_check')

    if not last_check:
        persist['last_check'] = datetime.utcnow()
        return
    if datetime.utcnow() - last_check <= timedelta(hours=2):
        # It has been less than 2 hour, don't check again
        return

    new_last_check = datetime.utcnow()

    try:
        # Calculate seconds since epoch minus a minute for buffer
        last_check_epoch = int((last_check - datetime(1970, 1, 1)).total_seconds()) - 60
        logger.debug('Getting updates from thetvdb ({})', last_check_epoch)
        updates = TVDBRequest().get('updated/query', fromTime=last_check_epoch)
    except requests.RequestException as e:
        logger.error('Could not get update information from tvdb: {}', e)
        return

    expired_series = [series['id'] for series in updates] if updates else []

    # Update our cache to mark the items that have expired
    for chunk in chunked(expired_series):
        series_updated = (
            session.query(TVDBSeries)
            .filter(TVDBSeries.id.in_(chunk))
            .update({'expired': True}, 'fetch')
        )
        episodes_updated = (
            session.query(TVDBEpisode)
            .filter(TVDBEpisode.series_id.in_(chunk))
            .update({'expired': True}, 'fetch')
        )
        logger.debug(
            '{} series and {} episodes marked as expired', series_updated, episodes_updated
        )

    persist['last_check'] = new_last_check



import base64
import datetime

from loguru import logger
from requests.exceptions import RequestException

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.plugin import PluginWarning
from flexget.utils.requests import Session as RequestSession
from flexget.utils.requests import TimedLimiter

plugin_name = 'pushbullet'
logger = logger.bind(name=plugin_name)

PUSHBULLET_URL = 'https://api.pushbullet.com/v2/pushes'

requests = RequestSession(max_retries=3)
requests.add_domain_limiter(TimedLimiter('pushbullet.com', '5 seconds'))


class PushbulletNotifier:
    """
    Example::

    notify:
      entries:
        via:
          pushbullet:
            apikey: <API_KEY>
            [device: <DEVICE_IDEN> (can also be a list of device ids, or don't specify any ids to send to all devices)]
            [email: <EMAIL_ADDRESS> (can also be a list of user email addresses)]
            [channel: <CHANNEL_TAG> (you can only specify device / email or channel tag, cannot use both)]

    Configuration parameters are also supported from entries (eg. through set).
    """

    schema = {
        'type': 'object',
        'properties': {
            'api_key': one_or_more({'type': 'string'}),
            'device': one_or_more({'type': 'string'}),
            'email': one_or_more({'type': 'string', 'format': 'email'}),
            'url': {'type': 'string'},
            'channel': {'type': 'string'},
            'file_template': {'type': 'string'},
        },
        'required': ['api_key'],
        'oneOf': [
            {'required': ['device']},
            {'required': ['channel']},
            {'required': ['email']},
            {
                'not': {
                    'anyOf': [
                        {'required': ['device']},
                        {'required': ['channel']},
                        {'required': ['email']},
                    ]
                }
            },
        ],
        'error_oneOf': 'One (and only one) of `email`, `device` or `channel` are allowed.',
        'additionalProperties': False,
    }

    def notify(self, title, message, config):
        """
        Send a Pushbullet notification
        """
        if config.get('device') and not isinstance(config['device'], list):
            config['device'] = [config['device']]

        if config.get('email') and not isinstance(config['email'], list):
            config['email'] = [config['email']]

        if not isinstance(config['api_key'], list):
            config['api_key'] = [config['api_key']]

        for key in config['api_key']:
            if config.get('channel'):
                self.send_push(
                    key, title, message, config.get('url'), config.get('channel'), 'channel_tag'
                )
            elif config.get('device'):
                for d in config['device']:
                    self.send_push(key, title, message, config.get('url'), d, 'device_iden')
            elif config.get('email'):
                for e in config['email']:
                    self.send_push(key, title, message, config.get('url'), e, 'email')
            else:
                self.send_push(key, title, message, config.get('url'))

    @staticmethod
    def send_push(api_key, title, body, url=None, destination=None, destination_type=None):
        push_type = 'link' if url else 'note'
        notification = {'type': push_type, 'title': title, 'body': body}
        if url:
            notification['url'] = url
        if destination:
            notification[destination_type] = destination

        # Make the request
        headers = {
            'Authorization': b'Basic ' + base64.b64encode(api_key.encode('ascii')),
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'User-Agent': 'Flexget',
        }
        try:
            response = requests.post(PUSHBULLET_URL, headers=headers, json=notification)
        except RequestException as e:
            if e.response is not None:
                if e.response.status_code == 429:
                    reset_time = e.response.headers.get('X-Ratelimit-Reset')
                    if reset_time:
                        reset_time = datetime.datetime.fromtimestamp(int(reset_time)).strftime(
                            '%Y-%m-%d %H:%M:%S'
                        )
                        message = f'Monthly Pushbullet database operations limit reached. Next reset: {reset_time}'
                else:
                    message = e.response.json()['error']['message']
            else:
                message = str(e)
            raise PluginWarning(message)

        reset_time = response.headers.get('X-Ratelimit-Reset')
        remaining = response.headers.get('X-Ratelimit-Remaining')
        if reset_time and remaining:
            reset_time = datetime.datetime.fromtimestamp(int(reset_time))
            logger.debug(
                'Pushbullet notification sent. Database operations remaining until next reset: {}. '
                'Next reset at: {}',
                remaining,
                reset_time,
            )


@event('plugin.register')
def register_plugin():
    plugin.register(PushbulletNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



import pytest

from flexget.api.app import base_message
from flexget.components.seen.db import SeenEntry
from flexget.components.series.api import ObjectsContainer as OC

# TODO: would be nicer to import db module
from flexget.components.series.db import (
    AlternateNames,
    Episode,
    EpisodeRelease,
    Season,
    SeasonRelease,
    Series,
    SeriesTask,
)
from flexget.components.thetvdb.api import ObjectsContainer as tvdb
from flexget.components.tvmaze.api import ObjectsContainer as tvmaze
from flexget.manager import Session
from flexget.utils import json


class TestSeriesRootAPI:
    config = """
        tasks: {}
    """

    def test_series_root_get(self, api_client, schema_match):
        # No params
        rsp = api_client.get('/series/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert data == []

        with Session() as session:
            series = Series()
            series.name = 'test series'
            session.add(series)

            task = SeriesTask('test task')
            series.in_tasks = [task]

        rsp = api_client.get('/series/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        show = data[0]
        errors = schema_match(OC.single_series_object, show)
        assert not errors

        assert show['name'] == 'test series'

    def test_series_configured_param(self, api_client, schema_match):
        with Session() as session:
            series = Series()
            series.name = 'test series'
            session.add(series)

        # Default is configured series, no results
        rsp = api_client.get('/series/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert data == []

        # Get unconfigured series
        rsp = api_client.get('/series/?in_config=unconfigured')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        show = data[0]
        errors = schema_match(OC.single_series_object, show)
        assert not errors

        assert len(data) == 1
        assert show['name'] == 'test series'

        # Add a configured series
        with Session() as session:
            series = Series()
            series.name = 'test series 2'
            session.add(series)

            task = SeriesTask('test task')
            series.in_tasks = [task]

        # Get only configures series
        rsp = api_client.get('/series/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        show = data[0]
        errors = schema_match(OC.single_series_object, show)
        assert not errors

        assert len(data) == 1
        assert show['name'] == 'test series 2'

        # Get all series
        rsp = api_client.get('/series/?in_config=all')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert len(data) == 2

    def test_series_premieres_param(self, api_client, schema_match):
        # Add a series with an episode of S02E05, not a premiere
        with Session() as session:
            series = Series()
            series.name = 'test series'
            session.add(series)

            task = SeriesTask('test task')
            series.in_tasks = [task]

            episode = Episode()
            episode.identifier = 'S02E05'
            episode.identified_by = 'ep'
            episode.season = 2
            episode.number = 5
            episode.series_id = series.id
            series.episodes.append(episode)

            release = EpisodeRelease()
            release.title = 'test release'
            release.downloaded = True

            episode.releases = [release]

        # Default all, not just premieres
        rsp = api_client.get('/series/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert len(data) == 1

        # Get only premieres
        rsp = api_client.get('/series/?premieres=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert len(data) == 0

        # Add a premiere episode to another series
        with Session() as session:
            series = Series()
            series.name = 'test series 2'
            session.add(series)

            task = SeriesTask('test task 2')
            series.in_tasks = [task]

            episode = Episode()
            episode.identifier = 'S01E01'
            episode.identified_by = 'ep'
            episode.season = 1
            episode.number = 1
            series.episodes.append(episode)

            release = EpisodeRelease()
            release.title = 'test release 2'
            release.downloaded = True
            episode.releases = [release]

        # Get only just premieres
        rsp = api_client.get('/series/?premieres=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert len(data) == 1

    @pytest.mark.online
    def test_series_lookup_param(self, api_client, schema_match):
        # Add two real shows
        with Session() as session:
            series = Series()
            series.name = 'Suits'
            session.add(series)

            series2 = Series()
            series2.name = 'Stranger Things'
            session.add(series2)

        rsp = api_client.get('/series/?in_config=all&lookup=tvdb&lookup=tvmaze')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert len(data) == 2

        for show in data:
            tvdb_lookup = show['lookup']['tvdb']
            assert tvdb_lookup
            errors = schema_match(tvdb.tvdb_series_object, tvdb_lookup)
            assert not errors

            tvmaze_lookup = show['lookup']['tvmaze']
            assert tvmaze_lookup
            errors = schema_match(tvmaze.tvmaze_series_object, tvmaze_lookup)
            assert not errors

    def test_series_post(self, api_client, schema_match):
        payload = {'name': 'test series'}

        # Minimal payload
        rsp = api_client.json_post('/series/', data=json.dumps(payload))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.single_series_object, data)
        assert not errors

        # Try to add again
        rsp = api_client.json_post('/series/', data=json.dumps(payload))
        assert rsp.status_code == 409, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'name': 'test series 2', 'begin_episode': 'bla'}

        # Invalid begin episode
        rsp = api_client.json_post('/series/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload3 = {
            'name': 'test series 2',
            'begin_episode': 's01e01',
            'alternate_names': ['show1', 'show2'],
        }

        # Maximal payload
        rsp = api_client.json_post('/series/', data=json.dumps(payload3))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.single_series_object, data)
        assert not errors

        assert data['name'] == payload3['name']
        assert data['alternate_names'] == payload3['alternate_names']
        assert data['begin_episode']['identifier'].lower() == payload3['begin_episode']

        payload4 = {'name': 'test series 3', 'alternate_names': ['show1']}

        # Alternate name already added to different show
        rsp = api_client.json_post('/series/', data=json.dumps(payload4))
        assert rsp.status_code == 409, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors


class TestSeriesSearchAPI:
    config = """
        tasks: {}
    """

    def test_series_search(self, api_client, schema_match):
        with Session() as session:
            series1 = Series()
            series1.name = 'test series1'
            session.add(series1)

            series2 = Series()
            series2.name = 'test series2'
            session.add(series2)

        rsp = api_client.get('/series/search/test/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert len(data) == 2

        rsp = api_client.get('/series/search/series1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert len(data) == 1

        rsp = api_client.get('/series/search/bla/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert len(data) == 0


class TestSeriesSingleAPI:
    config = """
        tasks: {}
    """

    def test_series_get(self, api_client, schema_match):
        with Session() as session:
            series1 = Series()
            series1.name = 'test series1'
            session.add(series1)

        rsp = api_client.get('/series/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.single_series_object, data)
        assert not errors

        assert data['name'] == 'test series1'

        # No existing ID
        rsp = api_client.get('/series/10/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_series_delete(self, api_client, schema_match):
        with Session() as session:
            series1 = Series()
            series1.name = 'test series1'
            session.add(series1)

        rsp = api_client.delete('/series/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Delete again, no existing ID
        rsp = api_client.delete('/series/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_series_put(self, api_client, schema_match):
        with Session() as session:
            series1 = Series()
            series1.name = 'test series1'
            session.add(series1)

        payload = {}

        # Validation error
        rsp = api_client.json_put('/series/1/', data=json.dumps(payload))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload1 = {'begin_episode': 's01e01', 'alternate_names': ['show1']}

        rsp = api_client.json_put('/series/1/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.single_series_object, data)
        assert not errors

        assert data['begin_episode']['identifier'].lower() == payload1['begin_episode']
        assert data['alternate_names'] == payload1['alternate_names']

        with Session() as session:
            series = Series()
            series.name = 'test series2'
            session.add(series)

            alt = AlternateNames('show2')
            series.alternate_names = [alt]

        payload2 = {'alternate_names': ['show2']}

        # Alternate name used by another show
        rsp = api_client.json_put('/series/1/', data=json.dumps(payload2))
        assert rsp.status_code == 409, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Non existent show
        rsp = api_client.json_put('/series/10/', data=json.dumps(payload2))
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors


class TestSeriesSeasonsAPI:
    config = """
        tasks: {}
    """

    def test_seasons_get(self, api_client, schema_match):
        with Session() as session:
            series = Series()
            series.name = 'test series 1'
            session.add(series)

            task = SeriesTask('test task')
            series.in_tasks = [task]

            season1 = Season()
            season1.identifier = 'S01'
            season1.identified_by = 'ep'
            season1.season = 1
            season1.series_id = series.id

            season2 = Season()
            season2.identifier = 'S02'
            season2.identified_by = 'ep'
            season2.season = 2
            season2.series_id = series.id

            release = SeasonRelease()
            release.title = 'test release'
            release.downloaded = True

            season1.releases = [release]

            series.seasons.append(season1)
            series.seasons.append(season2)

        # No series
        rsp = api_client.get('/series/10/seasons/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/seasons/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.seasons_list_schema, data)
        assert not errors

        assert len(data) == 2

        # Delete all episodes
        rsp = api_client.delete('/series/1/seasons/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/seasons/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.seasons_list_schema, data)
        assert not errors

        assert len(data) == 0


class TestSeriesSeasonAPI:
    config = """
        tasks: {}
    """

    def test_season(self, api_client, schema_match):
        with Session() as session:
            series = Series()
            series.name = 'test series 1'
            session.add(series)

            task = SeriesTask('test task')
            series.in_tasks = [task]

            season1 = Season()
            season1.identifier = 'S01'
            season1.identified_by = 'ep'
            season1.season = 1
            season1.series_id = series.id

            season2 = Season()
            season2.identifier = 'S02'
            season2.identified_by = 'ep'
            season2.season = 2
            season2.series_id = series.id

            release = SeasonRelease()
            release.title = 'test release'
            release.downloaded = True

            season1.releases = [release]

            series.seasons.append(season1)
            series.seasons.append(season2)

            series2 = Series()
            series2.name = 'test series 2'
            session.add(series2)

        rsp = api_client.get('/series/1/seasons/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.season_object, data)
        assert not errors

        assert data['identifier'] == 'S01'
        assert data['identified_by'] == 'ep'
        assert data['season'] == 1

        # No series ID
        rsp = api_client.get('/series/10/seasons/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No season ID
        rsp = api_client.get('/series/1/seasons/10/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # season does not belong to series
        rsp = api_client.get('/series/2/seasons/1/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Delete
        rsp = api_client.delete('/series/1/seasons/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/seasons/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.delete('/series/1/seasons/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.delete('/series/10/seasons/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors


class TestSeriesSeasonReleasesAPI:
    config = """
        tasks: {}
    """

    def create_data(self):
        with Session() as session:
            series = Series()
            series.name = 'test series 1'
            session.add(series)

            season1 = Season()
            season1.identifier = 'S01'
            season1.identified_by = 'ep'
            season1.season = 1
            season1.series_id = series.id

            release1 = SeasonRelease()
            release1.title = 'downloaded release'
            release1.downloaded = True

            release2 = SeasonRelease()
            release2.title = 'un-downloaded release'
            release2.downloaded = False

            season1.releases = [release1, release2]
            series.seasons.append(season1)

            series2 = Series()
            series2.name = 'test series 2'
            session.add(series2)

            season2 = Season()
            season2.identifier = 'S02'
            season2.identified_by = 'ep'
            season2.season = 2
            season2.series_id = series2.id

            series2.seasons.append(season2)

    def test_season_releases_get(self, api_client, schema_match):
        self.create_data()

        rsp = api_client.get('/series/1/seasons/1/releases/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.season_release_list_schema, data)
        assert not errors

        assert len(data) == 2

        # Just downloaded releases
        rsp = api_client.get('/series/1/seasons/1/releases/?downloaded=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.season_release_list_schema, data)
        assert not errors

        assert len(data) == 1
        assert data[0]['title'] == 'downloaded release'

        # Just un-downloaded releases
        rsp = api_client.get('/series/1/seasons/1/releases/?downloaded=false')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.season_release_list_schema, data)
        assert not errors

        assert len(data) == 1
        assert data[0]['title'] == 'un-downloaded release'

        # No series
        rsp = api_client.get('/series/10/seasons/1/releases/?downloaded=false')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No episode for series
        rsp = api_client.get('/series/1/seasons/10/releases/?downloaded=false')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Episode does not belong to series
        rsp = api_client.get('/series/2/seasons/1/releases/?downloaded=false')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_releases_delete(self, api_client, schema_match):
        self.create_data()

        rsp = api_client.delete('/series/1/seasons/1/releases/?downloaded=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/seasons/1/releases/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.season_release_list_schema, data)
        assert not errors

        assert len(data) == 1
        assert data[0]['title'] == 'un-downloaded release'

        rsp = api_client.delete('/series/1/seasons/1/releases/?downloaded=false')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/seasons/1/releases/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.season_release_list_schema, data)
        assert not errors

        assert len(data) == 0

        # No series
        rsp = api_client.delete('/series/10/seasons/1/releases/?downloaded=false')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No episode for series
        rsp = api_client.delete('/series/1/seasons/10/releases/?downloaded=false')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Episode does not belong to series
        rsp = api_client.delete('/series/2/seasons/1/releases/?downloaded=false')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_releases_put(self, api_client, schema_match):
        self.create_data()

        rsp = api_client.json_put('/series/1/seasons/1/releases/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/seasons/1/releases/?downloaded=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.season_release_list_schema, data)
        assert not errors

        assert len(data) == 0

        rsp = api_client.get('/series/1/seasons/1/releases/?downloaded=false')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.season_release_list_schema, data)
        assert not errors

        assert len(data) == 2

        # No series
        rsp = api_client.json_put('/series/10/seasons/1/releases/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No episode for series
        rsp = api_client.json_put('/series/1/seasons/10/releases/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Episode does not belong to series
        rsp = api_client.json_put('/series/2/seasons/1/releases/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors


class TestSeriesEpisodesAPI:
    config = """
        tasks: {}
    """

    def test_episodes_get(self, api_client, schema_match):
        with Session() as session:
            series = Series()
            series.name = 'test series 1'
            session.add(series)

            task = SeriesTask('test task')
            series.in_tasks = [task]

            episode1 = Episode()
            episode1.identifier = 'S01E01'
            episode1.identified_by = 'ep'
            episode1.season = 1
            episode1.number = 1
            episode1.series_id = series.id

            episode2 = Episode()
            episode2.identifier = 'S01E01'
            episode2.identified_by = 'ep'
            episode2.season = 1
            episode2.number = 1
            episode2.series_id = series.id

            release = EpisodeRelease()
            release.title = 'test release'
            release.downloaded = True

            episode1.releases = [release]

            series.episodes.append(episode1)
            series.episodes.append(episode2)

        # No series
        rsp = api_client.get('/series/10/episodes/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/episodes/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_list_schema, data)
        assert not errors

        assert len(data) == 2

        # Delete all episodes
        rsp = api_client.delete('/series/1/episodes/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/episodes/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_list_schema, data)
        assert not errors

        assert len(data) == 0


class TestSeriesEpisodeAPI:
    config = """
        tasks: {}
    """

    def test_episode(self, api_client, schema_match):
        with Session() as session:
            series = Series()
            series.name = 'test series 1'
            session.add(series)

            task = SeriesTask('test task')
            series.in_tasks = [task]

            episode1 = Episode()
            episode1.identifier = 'S01E01'
            episode1.identified_by = 'ep'
            episode1.season = 1
            episode1.number = 1
            episode1.series_id = series.id

            episode2 = Episode()
            episode2.identifier = 'S01E01'
            episode2.identified_by = 'ep'
            episode2.season = 1
            episode2.number = 1
            episode2.series_id = series.id

            release = EpisodeRelease()
            release.title = 'test release'
            release.downloaded = True

            episode1.releases = [release]

            series.episodes.append(episode1)
            series.episodes.append(episode2)

            series2 = Series()
            series2.name = 'test series 2'
            session.add(series2)

        rsp = api_client.get('/series/1/episodes/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_object, data)
        assert not errors

        assert data['identifier'] == 'S01E01'
        assert data['identified_by'] == 'ep'
        assert data['season'] == 1
        assert data['number'] == 1
        assert data['premiere'] == 'Series Premiere'

        # No series ID
        rsp = api_client.get('/series/10/episodes/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No episode ID
        rsp = api_client.get('/series/1/episodes/10/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Episode does not belong to series
        rsp = api_client.get('/series/2/episodes/1/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Episode does not belong to series
        rsp = api_client.delete('/series/2/episodes/1/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Delete
        rsp = api_client.delete('/series/1/episodes/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/episodes/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.delete('/series/1/episodes/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.delete('/series/10/episodes/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors


class TestSeriesEpisodeReleasesAPI:
    config = """
        tasks: {}
    """

    def test_releases_get(self, api_client, schema_match):
        with Session() as session:
            series = Series()
            series.name = 'test series 1'
            session.add(series)

            episode1 = Episode()
            episode1.identifier = 'S01E01'
            episode1.identified_by = 'ep'
            episode1.season = 1
            episode1.number = 1
            episode1.series_id = series.id

            release1 = EpisodeRelease()
            release1.title = 'downloaded release'
            release1.downloaded = True

            release2 = EpisodeRelease()
            release2.title = 'un-downloaded release'
            release2.downloaded = False

            episode1.releases = [release1, release2]
            series.episodes.append(episode1)

            series2 = Series()
            series2.name = 'test series 2'
            session.add(series2)

            episode2 = Episode()
            episode2.identifier = 'S01E02'
            episode2.identified_by = 'ep'
            episode2.season = 1
            episode2.number = 2
            episode2.series_id = series2.id

            series2.episodes.append(episode2)

        rsp = api_client.get('/series/1/episodes/1/releases/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_release_list_schema, data)
        assert not errors

        assert len(data) == 2

        # Just downloaded releases
        rsp = api_client.get('/series/1/episodes/1/releases/?downloaded=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_release_list_schema, data)
        assert not errors

        assert len(data) == 1
        assert data[0]['title'] == 'downloaded release'

        # Just un-downloaded releases
        rsp = api_client.get('/series/1/episodes/1/releases/?downloaded=false')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_release_list_schema, data)
        assert not errors

        assert len(data) == 1
        assert data[0]['title'] == 'un-downloaded release'

        # No series
        rsp = api_client.get('/series/10/episodes/1/releases/?downloaded=false')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No episode for series
        rsp = api_client.get('/series/1/episodes/10/releases/?downloaded=false')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Episode does not belong to series
        rsp = api_client.get('/series/2/episodes/1/releases/?downloaded=false')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_releases_delete(self, api_client, schema_match):
        with Session() as session:
            series = Series()
            series.name = 'test series 1'
            session.add(series)

            episode1 = Episode()
            episode1.identifier = 'S01E01'
            episode1.identified_by = 'ep'
            episode1.season = 1
            episode1.number = 1
            episode1.series_id = series.id

            release1 = EpisodeRelease()
            release1.title = 'downloaded release'
            release1.downloaded = True

            release2 = EpisodeRelease()
            release2.title = 'un-downloaded release'
            release2.downloaded = False

            episode1.releases = [release1, release2]
            series.episodes.append(episode1)

            series2 = Series()
            series2.name = 'test series 2'
            session.add(series2)

            episode2 = Episode()
            episode2.identifier = 'S01E02'
            episode2.identified_by = 'ep'
            episode2.season = 1
            episode2.number = 2
            episode2.series_id = series2.id

            series2.episodes.append(episode2)

        rsp = api_client.delete('/series/1/episodes/1/releases/?downloaded=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/episodes/1/releases/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_release_list_schema, data)
        assert not errors

        assert len(data) == 1
        assert data[0]['title'] == 'un-downloaded release'

        rsp = api_client.delete('/series/1/episodes/1/releases/?downloaded=false')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/episodes/1/releases/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_release_list_schema, data)
        assert not errors

        assert len(data) == 0

        # No series
        rsp = api_client.delete('/series/10/episodes/1/releases/?downloaded=false')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No episode for series
        rsp = api_client.delete('/series/1/episodes/10/releases/?downloaded=false')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Episode does not belong to series
        rsp = api_client.delete('/series/2/episodes/1/releases/?downloaded=false')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_releases_put(self, api_client, schema_match):
        with Session() as session:
            series = Series()
            series.name = 'test series 1'
            session.add(series)

            episode1 = Episode()
            episode1.identifier = 'S01E01'
            episode1.identified_by = 'ep'
            episode1.season = 1
            episode1.number = 1
            episode1.series_id = series.id

            release1 = EpisodeRelease()
            release1.title = 'downloaded release'
            release1.downloaded = True

            release2 = EpisodeRelease()
            release2.title = 'un-downloaded release'
            release2.downloaded = False

            episode1.releases = [release1, release2]
            series.episodes.append(episode1)

            series2 = Series()
            series2.name = 'test series 2'
            session.add(series2)

            episode2 = Episode()
            episode2.identifier = 'S01E02'
            episode2.identified_by = 'ep'
            episode2.season = 1
            episode2.number = 2
            episode2.series_id = series2.id

            series2.episodes.append(episode2)

        rsp = api_client.json_put('/series/1/episodes/1/releases/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/episodes/1/releases/?downloaded=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_release_list_schema, data)
        assert not errors

        assert len(data) == 0

        rsp = api_client.get('/series/1/episodes/1/releases/?downloaded=false')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_release_list_schema, data)
        assert not errors

        assert len(data) == 2

        # No series
        rsp = api_client.json_put('/series/10/episodes/1/releases/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No episode for series
        rsp = api_client.json_put('/series/1/episodes/10/releases/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Episode does not belong to series
        rsp = api_client.json_put('/series/2/episodes/1/releases/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors


class TestSeriesEpisodeReleaseAPI:
    config = """
        tasks: {}
    """

    def test_release_get(self, api_client, schema_match):
        with Session() as session:
            series = Series()
            series.name = 'test series 1'
            session.add(series)

            episode1 = Episode()
            episode1.identifier = 'S01E01'
            episode1.identified_by = 'ep'
            episode1.season = 1
            episode1.number = 1
            episode1.series_id = series.id

            episode2 = Episode()
            episode2.identifier = 'S01E02'
            episode2.identified_by = 'ep'
            episode2.season = 1
            episode2.number = 2
            episode2.series_id = series.id

            release1 = EpisodeRelease()
            release1.title = 'downloaded release'
            release1.downloaded = True

            release2 = EpisodeRelease()
            release2.title = 'un-downloaded release'
            release2.downloaded = False

            release3 = EpisodeRelease()
            release3.title = 'downloaded release'
            release3.downloaded = True

            episode1.releases = [release1, release2]
            episode2.releases = [release3]
            series.episodes.append(episode1)
            series.episodes.append(episode2)

            series2 = Series()
            series2.name = 'test series 2'
            session.add(series2)

            episode2 = Episode()
            episode2.identifier = 'S01E02'
            episode2.identified_by = 'ep'
            episode2.season = 1
            episode2.number = 2
            episode2.series_id = series2.id

            series2.episodes.append(episode2)

        rsp = api_client.get('/series/1/episodes/1/releases/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_release_object, data)
        assert not errors

        assert data['downloaded'] is True
        assert data['title'] == 'downloaded release'

        rsp = api_client.get('/series/1/episodes/1/releases/2/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_release_object, data)
        assert not errors

        assert data['downloaded'] is False
        assert data['title'] == 'un-downloaded release'

        # No series
        rsp = api_client.get('/series/10/episodes/1/releases/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No episode for series
        rsp = api_client.get('/series/1/episodes/10/releases/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Episode does not belong to series
        rsp = api_client.get('/series/2/episodes/1/releases/1/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No release
        rsp = api_client.get('/series/1/episodes/1/releases/10/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Release does not belong to episode
        rsp = api_client.get('/series/1/episodes/1/releases/3/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_release_put(self, api_client, schema_match):
        with Session() as session:
            series = Series()
            series.name = 'test series 1'
            session.add(series)

            episode1 = Episode()
            episode1.identifier = 'S01E01'
            episode1.identified_by = 'ep'
            episode1.season = 1
            episode1.number = 1
            episode1.series_id = series.id

            episode2 = Episode()
            episode2.identifier = 'S01E02'
            episode2.identified_by = 'ep'
            episode2.season = 1
            episode2.number = 2
            episode2.series_id = series.id

            release1 = EpisodeRelease()
            release1.title = 'downloaded release'
            release1.downloaded = True

            release2 = EpisodeRelease()
            release2.title = 'un-downloaded release'
            release2.downloaded = False

            release3 = EpisodeRelease()
            release3.title = 'downloaded release'
            release3.downloaded = True

            episode1.releases = [release1, release2]
            episode2.releases = [release3]
            series.episodes.append(episode1)
            series.episodes.append(episode2)

            series2 = Series()
            series2.name = 'test series 2'
            session.add(series2)

            episode2 = Episode()
            episode2.identifier = 'S01E02'
            episode2.identified_by = 'ep'
            episode2.season = 1
            episode2.number = 2
            episode2.series_id = series2.id

            series2.episodes.append(episode2)

        # No series
        rsp = api_client.json_put('/series/10/episodes/1/releases/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.json_put('/series/1/episodes/1/releases/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_release_object, data)
        assert not errors

        # Cannot reset if already downloaded
        rsp = api_client.json_put('/series/1/episodes/1/releases/1/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No episode for series
        rsp = api_client.json_put('/series/1/episodes/10/releases/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Episode does not belong to series
        rsp = api_client.json_put('/series/2/episodes/1/releases/1/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No release
        rsp = api_client.json_put('/series/1/episodes/1/releases/10/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Release does not belong to episode
        rsp = api_client.json_put('/series/1/episodes/1/releases/3/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_release_delete(self, api_client, schema_match):
        with Session() as session:
            series = Series()
            series.name = 'test series 1'
            session.add(series)

            episode1 = Episode()
            episode1.identifier = 'S01E01'
            episode1.identified_by = 'ep'
            episode1.season = 1
            episode1.number = 1
            episode1.series_id = series.id

            episode2 = Episode()
            episode2.identifier = 'S01E02'
            episode2.identified_by = 'ep'
            episode2.season = 1
            episode2.number = 2
            episode2.series_id = series.id

            release1 = EpisodeRelease()
            release1.title = 'downloaded release'
            release1.downloaded = True

            release2 = EpisodeRelease()
            release2.title = 'un-downloaded release'
            release2.downloaded = False

            release3 = EpisodeRelease()
            release3.title = 'downloaded release'
            release3.downloaded = True

            episode1.releases = [release1, release2]
            episode2.releases = [release3]
            series.episodes.append(episode1)
            series.episodes.append(episode2)

            series2 = Series()
            series2.name = 'test series 2'
            session.add(series2)

            episode3 = Episode()
            episode3.identifier = 'S01E02'
            episode3.identified_by = 'ep'
            episode3.season = 1
            episode3.number = 2
            episode3.series_id = series2.id

            release4 = EpisodeRelease()
            release4.title = 'downloaded release'
            release4.downloaded = True

            episode3.releases = [release4]

            series2.episodes.append(episode3)

        # No series
        rsp = api_client.delete('/series/10/episodes/1/releases/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No episode for series
        rsp = api_client.delete('/series/1/episodes/10/releases/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Episode does not belong to series
        rsp = api_client.delete('/series/2/episodes/1/releases/1/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # No release
        rsp = api_client.delete('/series/1/episodes/1/releases/10/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Release does not belong to episode
        rsp = api_client.delete('/series/1/episodes/1/releases/3/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.delete('/series/1/episodes/1/releases/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Already deleted
        rsp = api_client.delete('/series/1/episodes/1/releases/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors


class TestSeriesForgetFlag:
    config = """
            tasks:
              series_data:
                mock:
                  - {title: 'series.foo.s01e01.720p.hdtv-flexget'}
                  - {title: 'series.foo.s01e01.1080p.hdtv-flexget'}
                  - {title: 'series.foo.s01e02.720p.hdtv-flexget'}
                  - {title: 'series.foo.s01e02.1080p.hdtv-flexget'}
                series:
                  - series foo:
                      qualities:
                        - 720p
                        - 1080p

    """

    def test_delete_series_with_forget_flag(self, execute_task, api_client, schema_match):
        task = execute_task('series_data')
        assert len(task.accepted) == 4

        # Get series
        rsp = api_client.get('/series/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert len(data) == 1

        # Get seen object
        with Session() as session:
            seen = session.query(SeenEntry).all()
            assert len(seen) == 4

        # Delete with forget flag
        rsp = api_client.delete('/series/1/?forget=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Get series
        rsp = api_client.get('/series/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.series_list_schema, data)
        assert not errors

        assert len(data) == 0

        # Get seen object
        with Session() as session:
            seen = session.query(SeenEntry).all()
            assert len(seen) == 0

    def test_delete_series_episode_with_forget_flag(self, execute_task, api_client, schema_match):
        task = execute_task('series_data')
        assert len(task.accepted) == 4

        # Get episode 1
        rsp = api_client.get('/series/1/episodes/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_object, data)
        assert not errors

        # Get episode 2
        rsp = api_client.get('/series/1/episodes/2/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_object, data)
        assert not errors

        # Get seen object
        with Session() as session:
            seen = session.query(SeenEntry).all()
            assert len(seen) == 4

        # Delete with forget flag
        rsp = api_client.delete('/series/1/episodes/1/?forget=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Get episode 1
        rsp = api_client.get('/series/1/episodes/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Get episode 2
        rsp = api_client.get('/series/1/episodes/2/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_object, data)
        assert not errors

        # Get seen object
        with Session() as session:
            seen = session.query(SeenEntry).all()
            assert len(seen) == 2

    def test_delete_series_release_with_forget_flag(self, execute_task, api_client, schema_match):
        task = execute_task('series_data')
        assert len(task.accepted) == 4

        # Get release 1 for episode 1
        rsp = api_client.get('/series/1/episodes/1/releases/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.episode_release_object, data)
        assert not errors

        # Get seen object
        with Session() as session:
            seen = session.query(SeenEntry).all()
            assert len(seen) == 4

        rsp = api_client.delete('/series/1/episodes/1/releases/1/?forget=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/series/1/episodes/1/releases/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        with Session() as session:
            seen = session.query(SeenEntry).all()
            assert len(seen) == 3


class TestSeriesPagination:
    config = 'tasks: {}'

    def test_series_pagination(self, api_client, link_headers):
        number_of_series = 200
        with Session() as session:
            for i in range(number_of_series):
                series = Series()
                session.add(series)

                series.name = 'test series {}'.format(i)
                task = SeriesTask('test task')
                series.in_tasks = [task]

        # Default values
        rsp = api_client.get('/series/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 50  # Default page size
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 50

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 2

        # Change page size
        rsp = api_client.get('/series/?per_page=100')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 100
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 100

        links = link_headers(rsp)
        assert links['last']['page'] == 2
        assert links['next']['page'] == 2

        # Get different page
        rsp = api_client.get('series/?page=2')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 50
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 50

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 3
        assert links['prev']['page'] == 1

    def test_episodes_pagination(self, api_client, link_headers):
        number_of_episodes = 200
        with Session() as session:
            series = Series()
            session.add(series)

            series.name = 'test series'
            task = SeriesTask('test task')
            series.in_tasks = [task]

            for i in range(number_of_episodes):
                episode = Episode()
                episode.identifier = 'S01E0{}'.format(i)
                episode.identified_by = 'ep'
                episode.season = 1
                episode.number = i
                episode.series_id = series.id
                series.episodes.append(episode)

        # Default values
        rsp = api_client.get('/series/1/episodes/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 50  # Default page size
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 50
        assert int(rsp.headers['Series-ID']) == 1

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 2

        # Change page size
        rsp = api_client.get('/series/1/episodes/?per_page=100')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 100
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 100
        assert int(rsp.headers['Series-ID']) == 1

        links = link_headers(rsp)
        assert links['last']['page'] == 2
        assert links['next']['page'] == 2

        # Get different page
        rsp = api_client.get('series/1/episodes/?page=2')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 50
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 50
        assert int(rsp.headers['Series-ID']) == 1

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 3
        assert links['prev']['page'] == 1

    def test_releases_pagination(self, api_client, link_headers):
        number_of_releases = 200
        with Session() as session:
            series = Series()
            session.add(series)

            series.name = 'test series'
            task = SeriesTask('test task')
            series.in_tasks = [task]

            episode = Episode()
            episode.identifier = 'S01E01'
            episode.identified_by = 'ep'
            episode.season = 1
            episode.number = 1
            episode.series_id = series.id
            series.episodes.append(episode)

            for i in range(number_of_releases):
                release = EpisodeRelease()
                release.title = 'test release {}'.format(i)
                release.downloaded = True

                episode.releases.append(release)

        # Default values
        rsp = api_client.get('/series/1/episodes/1/releases/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 50  # Default page size
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 50
        assert int(rsp.headers['Series-ID']) == 1
        assert int(rsp.headers['Episode-ID']) == 1

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 2

        # Change page size
        rsp = api_client.get('/series/1/episodes/1/releases/?per_page=100')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 100
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 100
        assert int(rsp.headers['Series-ID']) == 1
        assert int(rsp.headers['Episode-ID']) == 1

        links = link_headers(rsp)
        assert links['last']['page'] == 2
        assert links['next']['page'] == 2

        # Get different page
        rsp = api_client.get('series/1/episodes/1/releases/?page=2')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 50
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 50
        assert int(rsp.headers['Series-ID']) == 1
        assert int(rsp.headers['Episode-ID']) == 1

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 3
        assert links['prev']['page'] == 1



from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event

logger = logger.bind(name='disable')


def all_builtins():
    """Helper function to return an iterator over all builtin plugins."""
    return (p for p in plugin.plugins.values() if p.builtin)


class DisablePlugin:
    """
    Allows disabling built-ins, or plugins referenced by template/include plugin.

    Example::

      templates:
        movies:
          download: ~/torrents/movies/
          .
          .

      tasks:
        nzbs:
          template: movies
          disable:
            - download
          sabnzbd:
            .
            .

      # Task nzbs uses all other configuration from template movies but removes the download plugin
    """

    schema = one_or_more({'type': 'string'})
    disabled_builtins = None

    @plugin.priority(254)
    def on_task_start(self, task, config):
        disabled = set()

        if isinstance(config, str):
            config = [config]

        # Disable plugins explicitly included in config.
        for p in config:
            if p == 'builtins':
                continue
            try:
                task.disable_plugin(p)
            except ValueError:
                logger.error('Unknown plugin `{}`', p)
            else:
                disabled.add(p)

        # Disable all builtins mode.
        if 'builtins' in config:
            for p in all_builtins():
                task.disable_plugin(p.name)
                disabled.add(p.name)

        logger.debug('Disabled plugins: {}', ', '.join(disabled))


@event('plugin.register')
def register_plugin():
    plugin.register(DisablePlugin, 'disable', api_ver=2)



from collections.abc import Iterable

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.cached_input import cached

logger = logger.bind(name='from_imdb')


class FromIMDB:
    """
    This plugin enables generating entries based on an entity, an entity being a person, character or company.
    It's based on IMDBpy which is required (pip install imdbpy). The basic config required just an IMDB ID of the
    required entity.

    For example:

        from_imdb: ch0001354

    Schema description:
    Other than ID, all other properties are meant to filter the full list that the entity generates.

    id: string that relates to a supported entity type. For example: 'nm0000375'. Required.
    job_types: a string or list with job types from job_types. Default is 'actor'.
    content_types: A string or list with content types from content_types. Default is 'movie'.
    max_entries: The maximum number of entries that can return. This value's purpose is basically flood protection
        against unruly configurations that will return too many results. Default is 200.

    Advanced config example:
        dynamic_movie_queue:
            from_imdb:
              id: co0051941
              job_types:
                - actor
                - director
              content_types: tv series
            accept_all: yes
            movie_queue: add

    """

    job_types = [
        'actor',
        'actress',
        'director',
        'producer',
        'writer',
        'self',
        'editor',
        'miscellaneous',
        'editorial department',
        'cinematographer',
        'visual effects',
        'thanks',
        'music department',
        'in development',
        'archive footage',
        'soundtrack',
    ]

    content_types = [
        'movie',
        'tv series',
        'tv mini series',
        'video game',
        'video movie',
        'tv movie',
        'episode',
    ]

    content_type_conversion = {
        'movie': 'movie',
        'tv series': 'tv',
        'tv mini series': 'tv',
        'tv movie': 'tv',
        'episode': 'tv',
        'video movie': 'video',
        'video game': 'video game',
    }

    character_content_type_conversion = {
        'movie': 'feature',
        'tv series': 'tv',
        'tv mini series': 'tv',
        'tv movie': 'tv',
        'episode': 'tv',
        'video movie': 'video',
        'video game': 'video-game',
    }

    jobs_without_content_type = ['actor', 'actress', 'self', 'in development', 'archive footage']

    imdb_pattern = one_or_more(
        {
            'type': 'string',
            'pattern': r'(nm|co|ch)\d{7,8}',
            'error_pattern': 'Get the id from the url of the person/company you want to use,'
            ' e.g. http://imdb.com/text/<id here>/blah',
        },
        unique_items=True,
    )

    schema = {
        'oneOf': [
            imdb_pattern,
            {
                'type': 'object',
                'properties': {
                    'id': imdb_pattern,
                    'job_types': one_or_more(
                        {'type': 'string', 'enum': job_types}, unique_items=True
                    ),
                    'content_types': one_or_more(
                        {'type': 'string', 'enum': content_types}, unique_items=True
                    ),
                    'max_entries': {'type': 'integer'},
                    'match_type': {'type': 'string', 'enum': ['strict', 'loose']},
                },
                'required': ['id'],
                'additionalProperties': False,
            },
        ]
    }

    def prepare_config(self, config):
        """
        Converts config to dict form and sets defaults if needed
        """
        config = config
        if isinstance(config, str):
            config = {'id': [config]}
        elif isinstance(config, list):
            config = {'id': config}
        if isinstance(config, dict) and not isinstance(config['id'], list):
            config['id'] = [config['id']]

        config.setdefault('content_types', [self.content_types[0]])
        config.setdefault('job_types', [self.job_types[0]])
        config.setdefault('max_entries', 200)
        config.setdefault('match_type', 'strict')

        if isinstance(config.get('content_types'), str):
            logger.debug('Converted content type from string to list.')
            config['content_types'] = [config['content_types']]

        if isinstance(config['job_types'], str):
            logger.debug('Converted job type from string to list.')
            config['job_types'] = [config['job_types']]
        # Special case in case user meant to add actress instead of actor (different job types in IMDB)
        if 'actor' in config['job_types'] and 'actress' not in config['job_types']:
            config['job_types'].append('actress')

        return config

    def get_items(self, config):
        items = []
        for id in config['id']:
            try:
                entity_type, entity_object = self.get_entity_type_and_object(id)
            except Exception as e:
                logger.error(
                    'Could not resolve entity via ID: {}. '
                    'Either error in config or unsupported entity. Error:{}',
                    id,
                    e,
                )
                continue
            items += self.get_items_by_entity(
                entity_type,
                entity_object,
                config.get('content_types'),
                config.get('job_types'),
                config.get('match_type'),
            )
        return set(items)

    def get_entity_type_and_object(self, imdb_id):
        """
        Return a tuple of entity type and entity object
        :param imdb_id: string which contains IMDB id
        :return: entity type, entity object (person, company, etc.)
        """
        if imdb_id.startswith('nm'):
            person = self.ia.get_person(imdb_id[2:])
            logger.info('Starting to retrieve items for person: {}', person)
            return 'Person', person
        elif imdb_id.startswith('co'):
            company = self.ia.get_company(imdb_id[2:])
            logger.info('Starting to retrieve items for company: {}', company)
            return 'Company', company
        elif imdb_id.startswith('ch'):
            character = self.ia.get_character(imdb_id[2:])
            logger.info('Starting to retrieve items for Character: {}', character)
            return 'Character', character

    def get_items_by_entity(
        self, entity_type, entity_object, content_types, job_types, match_type
    ):
        """
        Gets entity object and return movie list using relevant method
        """
        if entity_type == 'Company':
            return self.items_by_company(entity_object)

        if entity_type == 'Character':
            return self.items_by_character(entity_object, content_types, match_type)

        elif entity_type == 'Person':
            return self.items_by_person(entity_object, job_types, content_types, match_type)

    def flatten_list(self, _list):
        """
        Gets a list of lists and returns a flat list
        """
        for el in _list:
            if isinstance(el, Iterable) and not isinstance(el, str):
                for sub in self.flatten_list(el):
                    yield sub
            else:
                yield el

    def flat_list(self, non_flat_list, remove_none=False):
        flat_list = self.flatten_list(non_flat_list)
        if remove_none:
            flat_list = [_f for _f in flat_list if _f]
        return flat_list

    def filtered_items(self, unfiltered_items, content_types, match_type):
        items = []
        unfiltered_items = set(unfiltered_items)
        for item in sorted(unfiltered_items):
            if match_type == 'strict':
                logger.debug(
                    'Match type is strict, verifying item type to requested content types'
                )
                self.ia.update(item)
                if item['kind'] in content_types:
                    logger.verbose(
                        'Adding item "{}" to list. Item kind is "{}"', item, item['kind']
                    )
                    items.append(item)
                else:
                    logger.verbose('Rejecting item "{}". Item kind is "{}', item, item['kind'])
            else:
                logger.debug('Match type is loose, all items are being added')
                items.append(item)
        return items

    def items_by_person(self, person, job_types, content_types, match_type):
        """
        Return item list for a person object
        """
        unfiltered_items = self.flat_list(
            [self.items_by_job_type(person, job_type, content_types) for job_type in job_types],
            remove_none=True,
        )

        return self.filtered_items(unfiltered_items, content_types, match_type)

    def items_by_content_type(self, person, job_type, content_type):
        return [
            _f
            for _f in (person.get(job_type + ' ' + self.content_type_conversion[content_type], []))
            if _f
        ]

    def items_by_job_type(self, person, job_type, content_types):
        items = (
            person.get(job_type, [])
            if job_type in self.jobs_without_content_type
            else [
                person.get(job_type + ' ' + 'documentary', [])
                and person.get(job_type + ' ' + 'short', [])
                and self.items_by_content_type(person, job_type, content_type)
                if content_type == 'movie'
                else self.items_by_content_type(person, job_type, content_type)
                for content_type in content_types
            ]
        )
        return [_f for _f in items if _f]

    def items_by_character(self, character, content_types, match_type):
        """
        Return items list for a character object
        :param character: character object
        :param content_types: content types as defined in config
        :return:
        """
        unfiltered_items = self.flat_list(
            [
                character.get(self.character_content_type_conversion[content_type])
                for content_type in content_types
            ],
            remove_none=True,
        )

        return self.filtered_items(unfiltered_items, content_types, match_type)

    def items_by_company(self, company):
        """
        Return items list for a company object
        :param company: company object
        :return: company items list
        """
        return company.get('production companies')

    @cached('from_imdb', persist='2 hours')
    def on_task_input(self, task, config):
        try:
            from imdb import IMDb

            self.ia = IMDb()
        except ImportError:
            logger.error(
                'IMDBPY is required for this plugin. Please install using "pip install imdbpy"'
            )
            return

        entries = []
        config = self.prepare_config(config)
        items = self.get_items(config)
        if not items:
            logger.error('Could not get IMDB item list, check your configuration.')
            return
        for item in items:
            entry = Entry(
                title=item['title'],
                imdb_id='tt' + self.ia.get_imdbID(item),
                url='',
                imdb_url=self.ia.get_imdbURL(item),
            )

            if entry.isvalid():
                if entry not in entries:
                    entries.append(entry)
                    if entry and task.options.test:
                        logger.info("Test mode. Entry includes:")
                        for key, value in list(entry.items()):
                            logger.info('     {}: {}', key.capitalize(), value)
            else:
                logger.error('Invalid entry created? {}', entry)
        if len(entries) <= config.get('max_entries'):
            return entries
        else:
            logger.warning(
                'Number of entries ({}) exceeds maximum allowed value {}. '
                'Edit your filters or raise the maximum value by entering a higher "max_entries"',
                len(entries),
                config.get('max_entries'),
            )
            return


@event('plugin.register')
def register_plugin():
    plugin.register(FromIMDB, 'from_imdb', api_ver=2)



from pathlib import Path

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.utils.log import log_once
from flexget.utils.template import RenderError

try:
    # NOTE: Importing other plugins is discouraged!
    from flexget.components.parsing import parsers as plugin_parsers
except ImportError:
    raise plugin.DependencyError(issued_by=__name__, missing='parsers')

logger = logger.bind(name='exists_series')


class FilterExistsSeries:
    """
    Intelligent series aware exists rejecting.

    Example::

      exists_series: /storage/series/
    """

    schema = {
        'anyOf': [
            one_or_more({'type': 'string', 'format': 'path'}),
            {
                'type': 'object',
                'properties': {
                    'path': one_or_more({'type': 'string', 'format': 'path'}),
                    'allow_different_qualities': {
                        'enum': ['better', True, False],
                        'default': False,
                    },
                },
                'required': ['path'],
                'additionalProperties': False,
            },
        ]
    }

    def prepare_config(self, config):
        # if config is not a dict, assign value to 'path' key
        if not isinstance(config, dict):
            config = {'path': config}
        # if only a single path is passed turn it into a 1 element list
        if isinstance(config['path'], str):
            config['path'] = [config['path']]
        return config

    @plugin.priority(-1)
    def on_task_filter(self, task, config):
        if not task.accepted:
            logger.debug('Scanning not needed')
            return
        config = self.prepare_config(config)
        accepted_series = {}
        paths = set()
        for entry in task.accepted:
            if 'series_parser' in entry:
                if entry['series_parser'].valid:
                    accepted_series.setdefault(entry['series_parser'].name, []).append(entry)
                    for folder in config['path']:
                        try:
                            paths.add(entry.render(folder))
                        except RenderError as e:
                            logger.error('Error rendering path `{}`: {}', folder, e)
                else:
                    logger.debug('entry {} series_parser invalid', entry['title'])
        if not accepted_series:
            logger.warning(
                'No accepted entries have series information. exists_series cannot filter them'
            )
            return

        # scan through
        # For speed, only test accepted entries since our priority should be after everything is accepted.
        for series in accepted_series:
            # make new parser from parser in entry
            series_parser = accepted_series[series][0]['series_parser']
            for folder in paths:
                folder = Path(folder).expanduser()
                if not folder.is_dir():
                    logger.warning('Directory {} does not exist', folder)
                    continue

                for filename in folder.iterdir():
                    # run parser on filename data
                    try:
                        disk_parser = plugin.get('parsing', self).parse_series(
                            data=filename.name, name=series_parser.name
                        )
                    except plugin_parsers.ParseWarning as pw:
                        disk_parser = pw.parsed
                        log_once(pw.value, logger=logger)
                    if disk_parser.valid:
                        logger.debug('name {} is same series as {}', filename.name, series)
                        logger.debug('disk_parser.identifier = {}', disk_parser.identifier)
                        logger.debug('disk_parser.quality = {}', disk_parser.quality)
                        logger.debug('disk_parser.proper_count = {}', disk_parser.proper_count)

                        for entry in accepted_series[series]:
                            logger.debug(
                                'series_parser.identifier = {}', entry['series_parser'].identifier
                            )
                            if disk_parser.identifier != entry['series_parser'].identifier:
                                logger.trace('wrong identifier')
                                continue
                            logger.debug(
                                'series_parser.quality = {}', entry['series_parser'].quality
                            )
                            if config.get('allow_different_qualities') == 'better':
                                if entry['series_parser'].quality > disk_parser.quality:
                                    logger.trace('better quality')
                                    continue
                            elif config.get('allow_different_qualities'):
                                if disk_parser.quality != entry['series_parser'].quality:
                                    logger.trace('wrong quality')
                                    continue
                            logger.debug(
                                'entry parser.proper_count = {}',
                                entry['series_parser'].proper_count,
                            )
                            if disk_parser.proper_count >= entry['series_parser'].proper_count:
                                entry.reject('episode already exists')
                                continue
                            else:
                                logger.trace('new one is better proper, allowing')
                                continue


@event('plugin.register')
def register_plugin():
    plugin.register(FilterExistsSeries, 'exists_series', interfaces=['task'], api_ver=2)



from datetime import datetime, timedelta

from loguru import logger
from sqlalchemy import Column, DateTime, Integer, String, Unicode
from sqlalchemy.schema import Index

from flexget import db_schema
from flexget.event import event
from flexget.utils.sqlalchemy_utils import table_add_column

SCHEMA_VER = 3
FAIL_LIMIT = 100

logger = logger.bind(name='failed.db')
Base = db_schema.versioned_base('failed', SCHEMA_VER)


@db_schema.upgrade('failed')
def upgrade(ver, session):
    if ver is None or ver < 1:
        raise db_schema.UpgradeImpossible
    if ver == 1:
        table_add_column('failed', 'reason', Unicode, session)
        ver = 2
    if ver == 2:
        table_add_column('failed', 'retry_time', DateTime, session)
        ver = 3
    return ver


class FailedEntry(Base):
    __tablename__ = 'failed'

    id = Column(Integer, primary_key=True)
    title = Column(Unicode)
    url = Column(String)
    tof = Column(DateTime)
    reason = Column(Unicode)
    count = Column(Integer, default=1)
    retry_time = Column(DateTime)

    def __init__(self, title, url, reason=None):
        self.title = title
        self.url = url
        self.reason = reason
        self.tof = datetime.now()

    def __str__(self):
        return '<Failed(title=%s)>' % self.title

    def to_dict(self):
        return {
            'id': self.id,
            'title': self.title,
            'url': self.url,
            'added_at': self.tof,
            'reason': self.reason,
            'count': self.count,
            'retry_time': self.retry_time,
        }


# create indexes, used when creating tables
columns = Base.metadata.tables['failed'].c
Index('failed_title_url', columns.title, columns.url, columns.count)


@event('manager.db_cleanup')
def db_cleanup(manager, session):
    # Delete everything older than 30 days
    session.query(FailedEntry).filter(
        FailedEntry.tof < datetime.now() - timedelta(days=30)
    ).delete()
    # Of the remaining, always keep latest 25. Drop any after that if fail was more than a week ago.
    keep_num = 25
    keep_ids = [
        fe.id for fe in session.query(FailedEntry).order_by(FailedEntry.tof.desc())[:keep_num]
    ]
    if len(keep_ids) == keep_num:
        query = session.query(FailedEntry)
        query = query.filter(FailedEntry.id.notin_(keep_ids))
        query = query.filter(FailedEntry.tof < datetime.now() - timedelta(days=7))
        query.delete(synchronize_session=False)


def get_failures(session, count=None, start=None, stop=None, sort_by=None, descending=None):
    query = session.query(FailedEntry)
    if count:
        return query.count()
    if descending:
        query = query.order_by(getattr(FailedEntry, sort_by).desc())
    else:
        query = query.order_by(getattr(FailedEntry, sort_by))
    return query.slice(start, stop).all()



import pickle
from collections.abc import MutableSet
from datetime import datetime

from loguru import logger
from sqlalchemy import Column, DateTime, ForeignKey, Integer, Unicode, func, or_, select
from sqlalchemy.orm import relationship
from sqlalchemy.sql.elements import and_

from flexget import db_schema
from flexget.db_schema import versioned_base
from flexget.entry import Entry
from flexget.manager import Session
from flexget.utils import json, serialization
from flexget.utils.database import entry_synonym, with_session
from flexget.utils.sqlalchemy_utils import table_add_column, table_schema

logger = logger.bind(name='entry_list.db')
Base = versioned_base('entry_list', 2)


@db_schema.upgrade('entry_list')
def upgrade(ver, session):
    if None is ver:
        ver = 0
    if ver == 0:
        table = table_schema('entry_list_entries', session)
        table_add_column(table, 'json', Unicode, session)
        # Make sure we get the new schema with the added column
        table = table_schema('entry_list_entries', session)
        for row in session.execute(select([table.c.id, table.c.entry])):
            try:
                p = pickle.loads(row['entry'])
                session.execute(
                    table.update()
                    .where(table.c.id == row['id'])
                    .values(json=json.dumps(p, encode_datetime=True))
                )
            except KeyError as e:
                logger.error('Unable error upgrading entry_list pickle object due to {}', str(e))

        ver = 1
    if ver == 1:
        table = table_schema('entry_list_entries', session)
        for row in session.execute(select([table.c.id, table.c.json])):
            if not row['json']:
                # Seems there could be invalid data somehow. See #2590
                continue
            data = json.loads(row['json'], decode_datetime=True)
            # If title looked like a date, make sure it's a string
            title = str(data.pop('title'))
            e = Entry(title=title, **data)
            session.execute(
                table.update().where(table.c.id == row['id']).values(json=serialization.dumps(e))
            )
        ver = 2
    return ver


class EntryListList(Base):
    __tablename__ = 'entry_list_lists'
    id = Column(Integer, primary_key=True)
    name = Column(Unicode, unique=True)
    added = Column(DateTime, default=datetime.now)
    entries = relationship(
        'EntryListEntry', backref='list', cascade='all, delete, delete-orphan', lazy='dynamic'
    )

    def to_dict(self):
        return {'id': self.id, 'name': self.name, 'added_on': self.added}


class EntryListEntry(Base):
    __tablename__ = 'entry_list_entries'
    id = Column(Integer, primary_key=True)
    list_id = Column(Integer, ForeignKey(EntryListList.id), nullable=False)
    added = Column(DateTime, default=datetime.now)
    title = Column(Unicode)
    original_url = Column(Unicode)
    _json = Column('json', Unicode)
    entry = entry_synonym('_json')

    def __init__(self, entry, entry_list_id):
        self.title = entry['title']
        self.original_url = entry.get('original_url') or entry['url']
        self.entry = entry
        self.list_id = entry_list_id

    def __repr__(self):
        return '<EntryListEntry,title=%s,original_url=%s>' % (self.title, self.original_url)

    def to_dict(self):
        return {
            'id': self.id,
            'list_id': self.list_id,
            'added_on': self.added,
            'title': self.title,
            'original_url': self.original_url,
            'entry': json.coerce(self.entry),
        }


class DBEntrySet(MutableSet):
    def _db_list(self, session):
        return session.query(EntryListList).filter(EntryListList.name == self.config).first()

    def __init__(self, config):
        self.config = config
        with Session() as session:
            if not self._db_list(session):
                session.add(EntryListList(name=self.config))

    def _entry_query(self, session, entry):
        db_entry = (
            session.query(EntryListEntry)
            .filter(
                and_(
                    EntryListEntry.list_id == self._db_list(session).id,
                    or_(
                        EntryListEntry.title == entry['title'],
                        and_(
                            EntryListEntry.original_url,
                            EntryListEntry.original_url == entry['original_url'],
                        ),
                    ),
                )
            )
            .first()
        )

        return db_entry

    def __iter__(self):
        with Session() as session:
            for e in self._db_list(session).entries.order_by(EntryListEntry.added.desc()).all():
                logger.debug('returning {}', e.entry)
                yield e.entry

    def __contains__(self, entry):
        with Session() as session:
            return self._entry_query(session, entry) is not None

    def __len__(self):
        with Session() as session:
            return self._db_list(session).entries.count()

    def discard(self, entry):
        with Session() as session:
            db_entry = self._entry_query(session=session, entry=entry)
            if db_entry:
                logger.debug('deleting entry {}', db_entry)
                session.delete(db_entry)

    def add(self, entry):
        with Session() as session:
            stored_entry = self._entry_query(session, entry)
            if stored_entry:
                # Refresh all the fields if we already have this entry
                logger.debug('refreshing entry {}', entry)
                stored_entry.entry = entry
            else:
                logger.debug('adding entry {} to list {}', entry, self._db_list(session).name)
                stored_entry = EntryListEntry(entry=entry, entry_list_id=self._db_list(session).id)
            session.add(stored_entry)

    @property
    def immutable(self):
        return False

    def _from_iterable(self, it):
        # TODO: is this the right answer? the returned object won't have our custom __contains__ logic
        return set(it)

    @property
    def online(self):
        """Set the online status of the plugin, online plugin should be treated differently in certain situations,
        like test mode"""
        return False

    def get(self, entry):
        with Session() as session:
            match = self._entry_query(session=session, entry=entry)
            return Entry(match.entry) if match else None


@with_session
def get_entry_lists(name=None, session=None):
    logger.debug('retrieving entry lists')
    query = session.query(EntryListList)
    if name:
        logger.debug('searching for entry lists with name {}', name)
        query = query.filter(EntryListList.name.contains(name))
    return query.all()


@with_session
def get_list_by_exact_name(name, session=None):
    logger.debug('returning entry list with name {}', name)
    return (
        session.query(EntryListList).filter(func.lower(EntryListList.name) == name.lower()).one()
    )


@with_session
def get_list_by_id(list_id, session=None):
    logger.debug('fetching entry list with id {}', list_id)
    return session.query(EntryListList).filter(EntryListList.id == list_id).one()


@with_session
def delete_list_by_id(list_id, session=None):
    entry_list = get_list_by_id(list_id=list_id, session=session)
    if entry_list:
        logger.debug('deleting entry list with id {}', list_id)
        session.delete(entry_list)


@with_session
def get_entries_by_list_id(
    list_id,
    start=None,
    stop=None,
    order_by='title',
    descending=False,
    entry_ids=None,
    session=None,
):
    logger.debug('querying entries from entry list with id {}', list_id)
    query = session.query(EntryListEntry).filter(EntryListEntry.list_id == list_id)
    if entry_ids:
        query = query.filter(EntryListEntry.id.in_(entry_ids))
    if descending:
        query = query.order_by(getattr(EntryListEntry, order_by).desc())
    else:
        query = query.order_by(getattr(EntryListEntry, order_by))
    return query.slice(start, stop).all()


@with_session
def get_entry_by_title(list_id, title, session=None):
    entry_list = get_list_by_id(list_id=list_id, session=session)
    if entry_list:
        return (
            session.query(EntryListEntry)
            .filter(and_(EntryListEntry.title == title, EntryListEntry.list_id == list_id))
            .first()
        )


@with_session
def get_entry_by_id(list_id, entry_id, session=None):
    logger.debug('fetching entry with id {} from list id {}', entry_id, list_id)
    return (
        session.query(EntryListEntry)
        .filter(and_(EntryListEntry.id == entry_id, EntryListEntry.list_id == list_id))
        .one()
    )



from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event

logger = logger.bind(name='require_field')


class FilterRequireField:
    """
    Rejects entries without defined field.

    Example::

      require_field: imdb_url
    """

    schema = one_or_more({'type': 'string'})

    @plugin.priority(32)
    def on_task_filter(self, task, config):
        if isinstance(config, str):
            config = [config]
        for entry in task.entries:
            for field in config:
                if field not in entry:
                    entry.reject('Required field %s is not present' % field)
                    break
                if entry[field] is None:
                    entry.reject('Required field %s is `None`' % field)
                    break


@event('plugin.register')
def register_plugin():
    plugin.register(FilterRequireField, 'require_field', api_ver=2)
